# coding: utf-8
"""

"""

from http import HTTPStatus
import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms import BooleanField, StringField
from wtforms import validators

from .. import frontend
from ...logic import users, errors


class UserReadOnlyForm(FlaskForm):
    action = StringField(validators=[validators.AnyOf(['toggle_read_only'])])
    should_be_read_only = BooleanField()


class UserHiddenForm(FlaskForm):
    action = StringField(validators=[validators.AnyOf(['toggle_hidden'])])
    should_be_hidden = BooleanField()


class UserActiveForm(FlaskForm):
    action = StringField(validators=[validators.AnyOf(['toggle_active'])])
    should_be_active = BooleanField()


@frontend.route('/users/me')
@flask_login.login_required
def current_user_profile():
    return flask.redirect(flask.url_for('.user_profile', user_id=flask_login.current_user.id))


@frontend.route('/users/<int:user_id>', methods=['GET', 'POST'])
@flask_login.login_required
def user_profile(user_id):
    try:
        user = users.get_user(user_id)
    except errors.UserDoesNotExistError:
        return flask.abort(404)
    if flask_login.current_user.is_admin and not flask_login.current_user.is_readonly:
        user_read_only_form = UserReadOnlyForm()
        user_read_only_form.should_be_read_only.default = not user.is_readonly
        if user_read_only_form.validate_on_submit():
            users.set_user_readonly(user.id, user_read_only_form.should_be_read_only.data)
            if user_read_only_form.should_be_read_only.data:
                flask.flash('The user has been marked as read only.', 'success')
            else:
                flask.flash('The user has been unmarked as read only.', 'success')
            return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
        user_hidden_form = UserHiddenForm()
        user_hidden_form.should_be_hidden.default = not user.is_hidden
        if user_hidden_form.validate_on_submit():
            users.set_user_hidden(user.id, user_hidden_form.should_be_hidden.data)
            if user_hidden_form.should_be_hidden.data:
                flask.flash('The user has been marked as hidden.', 'success')
            else:
                flask.flash('The user has been unmarked as hidden.', 'success')
            return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
        user_active_form = UserActiveForm()
        user_active_form.should_be_active.default = not user.is_active
        if user_active_form.validate_on_submit():
            users.set_user_active(user.id, user_active_form.should_be_active.data)
            if user_active_form.should_be_active.data:
                flask.flash('The user has been activated.', 'success')
            else:
                flask.flash('The user has been deactivated.', 'success')
            return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
    elif flask.request.method.lower() == 'post':
        return flask.abort(HTTPStatus.METHOD_NOT_ALLOWED)
    else:
        user_read_only_form = None
        user_hidden_form = None
        user_active_form = None
    return flask.render_template(
        'profile.html',
        user_read_only_form=user_read_only_form,
        user_hidden_form=user_hidden_form,
        user_active_form=user_active_form,
        user=user
    )



# coding: utf-8
"""

"""

import flask
import flask_login

from .. import frontend
from .forms import ToggleFavoriteActionForm, ToggleFavoriteInstrumentForm
from ...logic.favorites import add_favorite_action, remove_favorite_action, get_user_favorite_action_ids, add_favorite_instrument, remove_favorite_instrument, get_user_favorite_instrument_ids
from ..utils import check_current_user_is_not_readonly


@frontend.route('/users/me/favorite_actions/', methods=['POST'])
@flask_login.login_required
def toggle_favorite_action():
    check_current_user_is_not_readonly()
    toggle_favorite_action_form = ToggleFavoriteActionForm()
    if toggle_favorite_action_form.validate():
        action_id = toggle_favorite_action_form.action_id.data
        user_id = flask_login.current_user.id
        if action_id not in get_user_favorite_action_ids(user_id=user_id):
            add_favorite_action(action_id=action_id, user_id=user_id)
            flask.flash('The action has been added to your favorites.', 'success')
        else:
            remove_favorite_action(action_id=action_id, user_id=user_id)
            flask.flash('The action has been removed from your favorites.', 'success')
    else:
        flask.flash('An error occurred while editing your favorite actions. Please try again.', 'error')
    return flask.redirect(flask.url_for(
        '.actions',
        sample_id=flask.request.args.get('sample_id', None),
        t=flask.request.args.get('t', None)
    ))


@frontend.route('/users/me/favorite_instruments/', methods=['POST'])
@flask_login.login_required
def toggle_favorite_instrument():
    check_current_user_is_not_readonly()
    toggle_favorite_instrument_form = ToggleFavoriteInstrumentForm()
    if toggle_favorite_instrument_form.validate():
        instrument_id = toggle_favorite_instrument_form.instrument_id.data
        user_id = flask_login.current_user.id
        if instrument_id not in get_user_favorite_instrument_ids(user_id=user_id):
            add_favorite_instrument(instrument_id=instrument_id, user_id=user_id)
            flask.flash('The instrument has been added to your favorites.', 'success')
        else:
            remove_favorite_instrument(instrument_id=instrument_id, user_id=user_id)
            flask.flash('The instrument has been removed from your favorites.', 'success')
    else:
        print(toggle_favorite_instrument_form.errors)
        flask.flash('An error occurred while editing your favorite instruments. Please try again.', 'error')
    return flask.redirect(flask.url_for('.instruments'))



# coding: utf-8
"""
RESTful API for SampleDB
"""

import flask
from flask_restful import Resource

from .authentication import multi_auth
from ...logic import errors, users

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def user_to_json(user: users.User):
    return {
        'user_id': user.id,
        'name': user.name,
        'orcid': user.orcid if user.orcid else None,
        'affiliation': user.affiliation if user.affiliation else None
    }


class User(Resource):
    @multi_auth.login_required
    def get(self, user_id: int):
        try:
            user = users.get_user(user_id=user_id)
        except errors.UserDoesNotExistError:
            return {
                "message": "user {} does not exist".format(user_id)
            }, 404
        return user_to_json(user)


class CurrentUser(Resource):
    @multi_auth.login_required
    def get(self):
        return user_to_json(flask.g.user)


class Users(Resource):
    @multi_auth.login_required
    def get(self):
        return [user_to_json(user) for user in users.get_users()]



# coding: utf-8
"""
Implementation of validate(instance, schema)
"""

import re
import datetime
import typing

from ...logic import actions, objects, datatypes, users
from ...models import ActionType
from ..errors import ObjectDoesNotExistError, ValidationError, ValidationMultiError, UserDoesNotExistError
from .utils import units_are_valid


def validate(instance: typing.Union[dict, list], schema: dict, path: typing.Optional[typing.List[str]] = None) -> None:
    """
    Validates the given instance using the given schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if path is None:
        path = []
    if not isinstance(schema, dict):
        raise ValidationError('invalid schema (must be dict)', path)
    if 'type' not in schema:
        raise ValidationError('invalid schema (must contain type)', path)
    if schema['type'] == 'array':
        return _validate_array(instance, schema, path)
    elif schema['type'] == 'object':
        return _validate_object(instance, schema, path)
    elif schema['type'] == 'text':
        return _validate_text(instance, schema, path)
    elif schema['type'] == 'datetime':
        return _validate_datetime(instance, schema, path)
    elif schema['type'] == 'bool':
        return _validate_bool(instance, schema, path)
    elif schema['type'] == 'quantity':
        return _validate_quantity(instance, schema, path)
    elif schema['type'] == 'sample':
        return _validate_sample(instance, schema, path)
    elif schema['type'] == 'measurement':
        return _validate_measurement(instance, schema, path)
    elif schema['type'] == 'object_reference':
        return _validate_object_reference(instance, schema, path)
    elif schema['type'] == 'tags':
        return _validate_tags(instance, schema, path)
    elif schema['type'] == 'hazards':
        return _validate_hazards(instance, schema, path)
    elif schema['type'] == 'user':
        return _validate_user(instance, schema, path)
    else:
        raise ValidationError('invalid type', path)


def _validate_array(instance: list, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given array schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, list):
        raise ValidationError('instance must be list', path)
    if 'minItems' in schema and len(instance) < schema['minItems']:
        raise ValidationError('expected at least {} items'.format(schema['minItems']), path)
    if 'maxItems' in schema and len(instance) > schema['maxItems']:
        raise ValidationError('expected at most {} items'.format(schema['maxItems']), path)
    errors = []
    for index, item in enumerate(instance):
        try:
            validate(item, schema['items'], path + [str(index)])
        except ValidationError as e:
            errors.append(e)
    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_hazards(instance: list, schema: dict, path: typing.List[str]) -> None:
    """
    Validate the given instance using the given GHS hazards schema and raise a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    if path != ['hazards']:
        raise ValidationError('GHS hazards must be a top-level entry named "hazards"', path)
    valid_keys = {'_type', 'hazards'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'hazards':
        raise ValidationError('expected _type "hazards"', path)
    if not isinstance(instance['hazards'], list):
        raise ValidationError('hazards must be list', path)
    errors = []
    hazards = []
    for index, item in enumerate(instance['hazards']):
        if not isinstance(item, int):
            errors.append(ValidationError('invalid hazard index type: {}'.format(type(item)), path + ['hazards', str(index)]))
        elif item in hazards:
            errors.append(ValidationError('duplicate hazard index: {}'.format(item), path + ['hazards', str(index)]))
        elif item < 1 or item > 9:
            errors.append(ValidationError('invalid hazard index: {}'.format(item), path + ['hazards', str(index)]))
        else:
            hazards.append(item)

    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_tags(instance: list, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given tags schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'tags'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'tags':
        raise ValidationError('expected _type "tags"', path)
    if not isinstance(instance['tags'], list):
        raise ValidationError('tags must be list', path)
    errors = []
    tags = []
    for index, item in enumerate(instance['tags']):
        if not isinstance(item, str):
            errors.append(ValidationError('invalid tag type: {}'.format(type(item)), path + ['tags', str(index)]))
        elif item in tags:
            errors.append(ValidationError('duplicate tag: {}'.format(item), path + ['tags', str(index)]))
        elif item.lower() != item:
            errors.append(ValidationError('tag not lowercase: {}'.format(item), path + ['tags', str(index)]))
        elif any(c not in 'abcdefghijklmnopqrstuvwxyz0123456789_-äöüß' for c in item):
            errors.append(ValidationError('tag contains invalid character: {}'.format(item), path + ['tags', str(index)]))
        else:
            tags.append(item)

    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_object(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    errors = []
    if 'required' in schema:
        for property_name in schema['required']:
            if property_name not in instance:
                errors.append(ValidationError('missing required property "{}"'.format(property_name), path + [property_name]))
    for property_name, property_value in instance.items():
        try:
            if property_name not in schema['properties']:
                raise ValidationError('unknown property "{}"'.format(property_name), path + [property_name])
            else:
                validate(property_value, schema['properties'][property_name], path + [property_name])
        except ValidationError as e:
            errors.append(e)
    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_text(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given text object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'text'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'text':
        raise ValidationError('expected _type "text"', path)
    if not isinstance(instance['text'], str):
        raise ValidationError('text must be str', path)
    choices = schema.get('choices', None)
    if choices and instance['text'] not in choices:
        raise ValidationError('The text must be one of {}.'.format(choices), path)
    min_length = schema.get('minLength', 0)
    max_length = schema.get('maxLength', None)
    if len(instance['text']) < min_length:
        raise ValidationError('The text must be at least {} characters long.'.format(min_length), path)
    if max_length is not None and len(instance['text']) > max_length:
        raise ValidationError('The text must be at most {} characters long.'.format(max_length), path)
    if 'pattern' in schema:
        if re.match(schema['pattern'], instance['text']) is None:
            raise ValidationError('The text must match the pattern: {}.'.format(schema['pattern']), path)


def _validate_datetime(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given datetime object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'utc_datetime'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'datetime':
        raise ValidationError('expected _type "datetime"', path)
    if not isinstance(instance['utc_datetime'], str):
        raise ValidationError('utc_datetime must be str', path)
    try:
        datetime.datetime.strptime(instance['utc_datetime'], '%Y-%m-%d %H:%M:%S')
    except ValueError:
        raise ValidationError('Please enter the date and time in the format: YYYY-MM-DD HH:MM:SS.', path)


def _validate_bool(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given boolean object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'value'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'bool':
        raise ValidationError('expected _type "bool"', path)
    if not isinstance(instance['value'], bool):
        raise ValidationError('value must be bool', path)


def _validate_quantity(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given quantity object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'units', 'dimensionality', 'magnitude_in_base_units'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'quantity':
        raise ValidationError('expected _type "quantity"', path)
    if not isinstance(instance['units'], str):
        raise ValidationError('units must be str', path)
    if not units_are_valid(instance['units']):
        raise ValidationError('Invalid/Unknown units', path)
    if not isinstance(instance['magnitude_in_base_units'], float) and not isinstance(instance['magnitude_in_base_units'], int):
        raise ValidationError('magnitude_in_base_units must be float or int', path)
    try:
        quantity = datatypes.Quantity(instance['magnitude_in_base_units'], units=instance['units'])
    except Exception:
        raise ValidationError('Unable to create quantity', path)
    if not isinstance(instance['dimensionality'], str):
        raise ValidationError('dimensionality must be str', path)
    try:
        schema_quantity = datatypes.Quantity(1.0, units=schema['units'])
    except Exception:
        raise ValidationError('Unable to create schema quantity', path)
    if quantity.dimensionality != schema_quantity.dimensionality:
        raise ValidationError('Invalid units, expected units for dimensionality "{}"'.format(str(schema_quantity.dimensionality)), path)
    if str(quantity.dimensionality) != instance['dimensionality']:
        raise ValidationError('Invalid dimensionality, expected "{}"'.format(str(schema_quantity.dimensionality)), path)


def _validate_sample(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given sample object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'object_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'sample':
        raise ValidationError('expected _type "sample"', path)
    if not isinstance(instance['object_id'], int):
        raise ValidationError('object_id must be int', path)
    try:
        sample = objects.get_object(object_id=instance['object_id'])
    except ObjectDoesNotExistError:
        raise ValidationError('object does not exist', path)
    action = actions.get_action(sample.action_id)
    if action.type_id != ActionType.SAMPLE_CREATION:
        raise ValidationError('object must be sample', path)


def _validate_measurement(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given measurement object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'object_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'measurement':
        raise ValidationError('expected _type "measurement"', path)
    if not isinstance(instance['object_id'], int):
        raise ValidationError('object_id must be int', path)
    try:
        measurement = objects.get_object(object_id=instance['object_id'])
    except ObjectDoesNotExistError:
        raise ValidationError('object does not exist', path)
    action = actions.get_action(measurement.action_id)
    if action.type_id != ActionType.MEASUREMENT:
        raise ValidationError('object must be measurement', path)


def _validate_user(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given user object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'user_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'user':
        raise ValidationError('expected _type "user"', path)
    if not isinstance(instance['user_id'], int):
        raise ValidationError('user_id must be int', path)
    try:
        users.get_user(user_id=instance['user_id'])
    except UserDoesNotExistError:
        raise ValidationError('user does not exist', path)


def _validate_object_reference(instance: dict, schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given instance using the given object reference object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'object_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if instance['_type'] != 'object_reference':
        raise ValidationError('expected _type "object_reference"', path)
    if not isinstance(instance['object_id'], int):
        raise ValidationError('object_id must be int', path)
    try:
        object = objects.get_object(object_id=instance['object_id'])
    except ObjectDoesNotExistError:
        raise ValidationError('object does not exist', path)
    if 'action_id' in schema and isinstance(schema['action_id'], int) and object.action_id != schema['action_id']:
        raise ValidationError('object has wrong action', path)
    if 'action_type_id' in schema and isinstance(schema['action_type_id'], int):
        action = actions.get_action(object.action_id)
        if action.type_id != schema['action_type_id']:
            raise ValidationError('object has wrong action type', path)



# coding: utf-8
"""
Tags or keywords are a method of organizing object. Internally, tags are part
of an objects data as text entries. In addition, however, the list of tags and
the number of objects each tag has been used for is stored in the database, so
that users can get tag suggestions and see which tags are used more to foster
a more unified vocabulary without strictly enforcing a whitelist of tags.
"""


import collections
import typing
from .. import db
from ..models import tags, objects


class Tag(collections.namedtuple('Tag', ['id', 'name', 'uses'])):
    """
    This class provides an immutable wrapper around models.tags.Tag.
    """

    def __new__(cls, id: int, name: str, uses: int):
        self = super(Tag, cls).__new__(cls, id, name, uses)
        return self

    @classmethod
    def from_database(cls, tag: tags.Tag) -> 'Tag':
        return Tag(id=tag.id, name=tag.name, uses=tag.uses)


def get_object_tags(object: objects.Object) -> typing.Sequence[str]:
    if 'tags' in object.data:
        tags_entry = object.data['tags']
        if tags_entry['_type'] == 'tags' and 'tags' in tags_entry:
            return [tag.strip() for tag in tags_entry["tags"]]
    return ()


def update_object_tag_usage(object: objects.Object) -> None:
    previous_tags = ()
    new_tags = get_object_tags(object)
    if object.version_id > 0:
        previous_object = objects.Objects.get_object_version(object.object_id, object.version_id - 1)
        previous_tags = get_object_tags(previous_object)
    removed_tags = set(previous_tags) - set(new_tags)
    added_tags = set(new_tags) - set(previous_tags)
    for tag_name in added_tags:
        tag = tags.Tag.query.filter_by(name=tag_name).first()
        if tag is not None:
            tag.uses += 1
        else:
            tag = tags.Tag(name=tag_name, uses=1)
        db.session.add(tag)
    for tag_name in removed_tags:
        tag = tags.Tag.query.filter_by(name=tag_name).first()
        if tag is not None:
            tag.uses = max(0, tag.uses - 1)
            if tag.uses > 0:
                db.session.add(tag)
            else:
                db.session.delete(tag)
    db.session.commit()


def get_tags() -> typing.Sequence[Tag]:
    return [Tag.from_database(tag) for tag in tags.Tag.query.order_by(tags.Tag.uses).all()]



# coding: utf-8
"""
Script for setting whether a user is limited to READ permissions or not.

Usage: python -m sampledb set_user_readonly <user_id> <yes_or_no>
"""

import sys
from .. import create_app
from ..logic.users import get_user, set_user_readonly
from ..logic.errors import UserDoesNotExistError


def main(arguments):
    if len(arguments) != 2 or not all(arguments) or arguments[1] not in ('yes', 'no'):
        print(__doc__)
        exit(1)
    user_id, yes_or_no = arguments
    try:
        user_id = int(user_id)
    except ValueError:
        print("Error: user_id must be an integer", file=sys.stderr)
        exit(1)
    is_readonly = yes_or_no == 'yes'
    app = create_app()
    with app.app_context():
        try:
            user = get_user(user_id)
        except UserDoesNotExistError:
            print("Error: No user with this ID exists", file=sys.stderr)
            exit(1)
        if user.is_readonly == is_readonly:
            print("Success: no change necessary")
        else:
            set_user_readonly(user.id, is_readonly)
            if is_readonly:
                print("Success: the user is readonly")
            else:
                print("Success: the user is not readonly")



# coding: utf-8
"""
Script for updating an action in SampleDB.

Usage: python -m sampledb update_action <action_id> <name> <description> <schema_file_name>
"""

import json
import sys
from .. import create_app
from ..logic.actions import update_action, get_action
from ..logic.schemas import validate_schema
from ..logic.errors import ActionDoesNotExistError, ValidationError


def main(arguments):
    if len(arguments) != 4:
        print(__doc__)
        exit(1)
    action_id, name, description, schema_file_name = arguments
    try:
        action_id = int(action_id)
    except ValueError:
        print("Error: action_id must be an integer", file=sys.stderr)
        exit(1)

    app = create_app()
    with app.app_context():
        try:
            action = get_action(action_id)
        except ActionDoesNotExistError:
            print('Error: no action with this id exists', file=sys.stderr)
            exit(1)
        with open(schema_file_name, 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        try:
            validate_schema(schema)
        except ValidationError as e:
            print('Error: invalid schema: {}'.format(str(e)), file=sys.stderr)
            exit(1)
        update_action(
            action_id=action.id,
            name=name,
            description=description,
            schema=schema
        )
        print("Success: the action has been updated in SampleDB")




import flask
from werkzeug.middleware.proxy_fix import ProxyFix
from flask_login import LoginManager
from flask_mail import Mail
from flask_sqlalchemy import SQLAlchemy

login_manager = LoginManager()
login_manager.session_protection = 'basic'

mail = Mail()
db = SQLAlchemy()

import sampledb.frontend
import sampledb.api
import sampledb.logic
import sampledb.models
import sampledb.models.migrations
import sampledb.config


def setup_database(app):
    with app.app_context():
        db.metadata.create_all(bind=db.engine)
        sampledb.models.Objects.bind = db.engine
        sampledb.models.migrations.run(db)


def setup_admin_account_from_config(app):
    with app.app_context():
        if 'ADMIN_INFO' in app.config['internal'] and not sampledb.logic.users.get_users(exclude_hidden=False):
            admin_username, admin_email, admin_password = app.config['internal']['ADMIN_INFO']
            admin_user = sampledb.logic.users.create_user(
                admin_username,
                admin_email,
                sampledb.logic.users.UserType.PERSON
            )
            sampledb.logic.authentication.add_other_authentication(
                admin_user.id,
                admin_username,
                admin_password,
                confirmed=True
            )
            sampledb.logic.users.set_user_administrator(admin_user.id, True)


def setup_jinja_environment(app):
    with app.app_context():
        is_ldap_configured = sampledb.logic.ldap.is_ldap_configured()

    if app.config['JUPYTERHUB_TEMPLATES_URL']:
        jupyterhub_templates_url = app.config['JUPYTERHUB_TEMPLATES_URL']
    elif app.config['JUPYTERHUB_URL']:
        jupyterhub_templates_url = app.config['JUPYTERHUB_URL'] + '/templates'
    else:
        jupyterhub_templates_url = None

    app.jinja_env.globals.update(
        export_file_formats=sampledb.logic.export.FILE_FORMATS,
        jupyterhub_name=app.config['JUPYTERHUB_NAME'],
        jupyterhub_templates_url=jupyterhub_templates_url,
        signout_form=sampledb.frontend.users_forms.SignoutForm,
        service_name=app.config['SERVICE_NAME'],
        service_description=app.config['SERVICE_DESCRIPTION'],
        service_imprint=app.config['SERVICE_IMPRINT'],
        service_privacy_policy=app.config['SERVICE_PRIVACY_POLICY'],
        ldap_name=app.config['LDAP_NAME'],
        is_ldap_configured=is_ldap_configured,
        get_action_types=sampledb.logic.actions.get_action_types,
        contact_email=app.config['CONTACT_EMAIL']
    )
    app.jinja_env.filters.update(sampledb.frontend.utils.jinja_filter.filters)


def create_app():
    app = flask.Flask(__name__)
    app.wsgi_app = ProxyFix(app.wsgi_app)

    app.config.from_object(sampledb.config)

    internal_config = sampledb.config.check_config(app.config)
    app.config['internal'] = internal_config

    login_manager.init_app(app)
    mail.init_app(app)
    db.init_app(app)
    sampledb.api.server.api.init_app(app)

    app.register_blueprint(sampledb.frontend.frontend)

    login_manager.login_view = 'frontend.sign_in'

    sampledb.logic.files.FILE_STORAGE_PATH = app.config['FILE_STORAGE_PATH']

    setup_database(app)
    setup_admin_account_from_config(app)
    setup_jinja_environment(app)

    return app



# coding: utf-8
"""

"""

import enum
import datetime

from .. import db

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@enum.unique
class HTTPMethod(enum.Enum):
    GET = 0
    POST = 1
    PUT = 2
    PATCH = 3
    DELETE = 4
    HEAD = 5
    OPTIONS = 6
    OTHER = 7


class APILogEntry(db.Model):
    __tablename__ = 'api_log_entries'

    id = db.Column(db.Integer, primary_key=True)
    api_token_id = db.Column(db.Integer, db.ForeignKey('authentications.id'), nullable=False)
    method = db.Column(db.Enum(HTTPMethod), nullable=False)
    route = db.Column(db.String, nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)
    api_token = db.relationship('Authentication', backref=db.backref("api_log_entries", cascade="all,delete"))

    def __init__(self, api_token_id, method, route, utc_datetime=None):
        self.api_token_id = api_token_id
        self.method = method
        self.route = route
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        self.utc_datetime = utc_datetime

    def __repr__(self):
        return '<{0}(id={1.id}, api_token_id={1.api_token_id}, method={1.method}, route={1.route}, utc_datetime={1.utc_datetime})>'.format(type(self).__name__, self)



# coding: utf-8
"""
Logic module for instrument log entries

Instrument scientists can add messages to a log for their instruments. They
can also allow users to provide log entries, e.g. to report issues. These log
entries are immutable and therefore this module only allows the creation and
querying of comments.

Files can be attached to these log entries, e.g. to supply the instrument scientists
with control files or photographs of an experiment setup.
"""

import collections
import datetime
import typing

from .. import db
from . import errors, instruments, users, user_log, objects
from .notifications import create_notification_for_a_new_instrument_log_entry, create_notification_for_an_edited_instrument_log_entry
from ..models import instrument_log_entries
from ..models.instrument_log_entries import InstrumentLogCategoryTheme


class InstrumentLogEntry(
    collections.namedtuple(
        'InstrumentLogEntry',
        ['id', 'instrument_id', 'user_id', 'versions']
    )
):
    """
    This class provides an immutable wrapper around models.instrument_log_entries.InstrumentLogEntry.
    """

    def __new__(
            cls,
            id: int,
            instrument_id: int,
            user_id: int,
            versions: typing.Tuple['InstrumentLogEntryVersion'] = ()
    ):
        self = super(InstrumentLogEntry, cls).__new__(
            cls, id, instrument_id, user_id, versions
        )
        return self

    @classmethod
    def from_database(cls, instrument_log_entry: instrument_log_entries.InstrumentLogEntry) -> 'InstrumentLogEntry':
        return InstrumentLogEntry(
            id=instrument_log_entry.id,
            instrument_id=instrument_log_entry.instrument_id,
            user_id=instrument_log_entry.user_id,
            versions=tuple(sorted([
                InstrumentLogEntryVersion.from_database(version)
                for version in instrument_log_entry.versions
            ], key=lambda v: (v.utc_datetime, v.version_id)))
        )

    @property
    def author(self):
        return users.get_user(self.user_id)

    @property
    def file_attachments(self) -> typing.List['InstrumentLogFileAttachment']:
        return get_instrument_log_file_attachments(self.id)

    @property
    def object_attachments(self) -> typing.List['InstrumentLogObjectAttachment']:
        return get_instrument_log_object_attachments(self.id)


class InstrumentLogEntryVersion(
    collections.namedtuple(
        'InstrumentLogEntryVersion',
        ['log_entry_id', 'version_id', 'content', 'utc_datetime', 'categories']
    )
):
    """
    This class provides an immutable wrapper around models.instrument_log_entries.InstrumentLogEntryVersion.
    """

    def __new__(
            cls,
            log_entry_id: int,
            version_id: int,
            content: str,
            utc_datetime: datetime.datetime,
            categories: typing.List['InstrumentLogCategory']
    ):
        self = super(InstrumentLogEntryVersion, cls).__new__(
            cls, log_entry_id, version_id, content, utc_datetime, categories
        )
        return self

    @classmethod
    def from_database(cls, instrument_log_entry_version: instrument_log_entries.InstrumentLogEntryVersion) -> 'InstrumentLogEntryVersion':
        return InstrumentLogEntryVersion(
            log_entry_id=instrument_log_entry_version.log_entry_id,
            version_id=instrument_log_entry_version.version_id,
            content=instrument_log_entry_version.content,
            utc_datetime=instrument_log_entry_version.utc_datetime,
            categories=[
                InstrumentLogCategory.from_database(category)
                for category in sorted(instrument_log_entry_version.categories, key=lambda category: category.theme.value)
            ]
        )


class InstrumentLogFileAttachment(
    collections.namedtuple(
        'InstrumentLogFileAttachment',
        ['id', 'log_entry_id', 'file_name', 'content', 'is_hidden']
    )
):
    """
    This class provides an immutable wrapper around models.instrument_log_entries.InstrumentLogFileAttachment.
    """

    def __new__(
            cls,
            id: int,
            log_entry_id: int,
            file_name: str,
            content: bytes,
            is_hidden: bool = False
    ):
        self = super(InstrumentLogFileAttachment, cls).__new__(
            cls, id, log_entry_id, file_name, content, is_hidden
        )
        return self

    @classmethod
    def from_database(
            cls,
            instrument_log_file_attachment: instrument_log_entries.InstrumentLogFileAttachment
    ) -> 'InstrumentLogFileAttachment':
        return InstrumentLogFileAttachment(
            id=instrument_log_file_attachment.id,
            log_entry_id=instrument_log_file_attachment.log_entry_id,
            file_name=instrument_log_file_attachment.file_name,
            content=instrument_log_file_attachment.content,
            is_hidden=instrument_log_file_attachment.is_hidden
        )


class InstrumentLogObjectAttachment(
    collections.namedtuple(
        'InstrumentLogObjectAttachment',
        ['id', 'log_entry_id', 'object_id', 'is_hidden']
    )
):
    """
    This class provides an immutable wrapper around models.instrument_log_entries.InstrumentLogObjectAttachment.
    """

    def __new__(
            cls,
            id: int,
            log_entry_id: int,
            object_id: int,
            is_hidden: bool = False
    ):
        self = super(InstrumentLogObjectAttachment, cls).__new__(
            cls, id, log_entry_id, object_id, is_hidden
        )
        return self

    @classmethod
    def from_database(
            cls,
            instrument_log_object_attachment: instrument_log_entries.InstrumentLogObjectAttachment
    ) -> 'InstrumentLogObjectAttachment':
        return InstrumentLogObjectAttachment(
            id=instrument_log_object_attachment.id,
            log_entry_id=instrument_log_object_attachment.log_entry_id,
            object_id=instrument_log_object_attachment.object_id,
            is_hidden=instrument_log_object_attachment.is_hidden
        )


class InstrumentLogCategory(
    collections.namedtuple(
        'InstrumentLogCategory',
        ['id', 'instrument_id', 'title', 'theme']
    )
):
    """
    This class provides an immutable wrapper around models.instrument_log_entries.InstrumentLogCategory.
    """

    def __new__(
            cls,
            id: int,
            instrument_id: int,
            title: str,
            theme: instrument_log_entries.InstrumentLogCategoryTheme
    ):
        self = super(InstrumentLogCategory, cls).__new__(
            cls, id, instrument_id, title, theme
        )
        return self

    @classmethod
    def from_database(
            cls,
            category: instrument_log_entries.InstrumentLogCategory
    ) -> 'InstrumentLogCategory':
        return InstrumentLogCategory(
            id=category.id,
            instrument_id=category.instrument_id,
            title=category.title,
            theme=category.theme
        )


def get_instrument_log_categories(instrument_id: int) -> typing.List[InstrumentLogCategory]:
    """
    Return the instrument log categories for an object.

    :param instrument_id: the ID of an existing instrument
    :return: the instrument log categories
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    categories = instrument_log_entries.InstrumentLogCategory.query.filter_by(instrument_id=instrument_id).all()
    if not categories:
        # ensure that the instrument exists
        instruments.get_instrument(instrument_id=instrument_id)
        return []
    return [
        InstrumentLogCategory.from_database(category)
        for category in categories
    ]


def create_instrument_log_category(
    instrument_id: int,
    title: str,
    theme: instrument_log_entries.InstrumentLogCategoryTheme
) -> InstrumentLogCategory:
    """
    Create a new instrument log category for an instrument.

    :param instrument_id: the ID of an existing instrument
    :param title: the category title
    :param theme: the category theme
    :return: the newly created category
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    category = instrument_log_entries.InstrumentLogCategory(
        instrument_id=instrument_id,
        title=title,
        theme=theme
    )
    db.session.add(category)
    db.session.commit()
    return InstrumentLogCategory.from_database(category)


def update_instrument_log_category(
    category_id: int,
    title: str,
    theme: InstrumentLogCategoryTheme
) -> InstrumentLogCategory:
    """

    :param category_id:
    :param title:
    :param theme:
    :return: the updated log category
    :raise errors.InstrumentLogCategoryDoesNotExistError: when no log category
        with the given ID exists
    """
    category = instrument_log_entries.InstrumentLogCategory.query.filter_by(id=category_id).first()
    if category is None:
        raise errors.InstrumentLogCategoryDoesNotExistError()
    category.title = title
    category.theme = theme
    db.session.add(category)
    db.session.commit()
    return InstrumentLogCategory.from_database(category)


def delete_instrument_log_category(category_id: int) -> None:
    category = instrument_log_entries.InstrumentLogCategory.query.filter_by(id=category_id).first()
    if category is None:
        raise errors.InstrumentLogCategoryDoesNotExistError()
    db.session.delete(category)
    db.session.commit()


def create_instrument_log_entry(
        instrument_id: int,
        user_id: int,
        content: str,
        category_ids: typing.Sequence[int] = ()
) -> InstrumentLogEntry:
    """
    Create a new instrument log entry and adds it to the user log.

    :param instrument_id: the ID of an existing instrument
    :param user_id: the ID of an existing user
    :param content: the text content for the new log entry
    :param category_ids: the instrument-specific log category IDs
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.InstrumentLogCategoryDoesNotExistError: when no log category
        with the given ID exists for this instrument
    """
    # ensure that the instrument exists
    instrument = instruments.get_instrument(instrument_id=instrument_id)
    # ensure that the user exists
    users.get_user(user_id)
    categories = []
    for category_id in set(category_ids):
        category = instrument_log_entries.InstrumentLogCategory.query.filter_by(
            id=category_id,
            instrument_id=instrument_id
        ).first()
        if category:
            categories.append(category)
        else:
            raise errors.InstrumentLogCategoryDoesNotExistError()
    instrument_log_entry = instrument_log_entries.InstrumentLogEntry(
        instrument_id=instrument_id,
        user_id=user_id
    )
    db.session.add(instrument_log_entry)
    db.session.commit()
    instrument_log_entry_version = instrument_log_entries.InstrumentLogEntryVersion(
        log_entry_id=instrument_log_entry.id,
        version_id=1,
        content=content
    )
    instrument_log_entry_version.categories = list(categories)
    db.session.add(instrument_log_entry_version)
    db.session.commit()
    user_log.create_instrument_log_entry(
        user_id=user_id,
        instrument_id=instrument_id,
        instrument_log_entry_id=instrument_log_entry.id
    )
    for responsible_user in instrument.responsible_users:
        if responsible_user.id != user_id:
            create_notification_for_a_new_instrument_log_entry(
                user_id=responsible_user.id,
                instrument_log_entry_id=instrument_log_entry.id
            )
    return InstrumentLogEntry.from_database(instrument_log_entry)


def update_instrument_log_entry(
        log_entry_id: int,
        content: str,
        category_ids: typing.Sequence[int] = ()
) -> InstrumentLogEntry:
    """
    Update an existing instrument log entry.

    :param log_entry_id: the ID of an existing instrument log entry
    :param content: the text content for the new log entry
    :param category_ids: the instrument-specific log category IDs
    :raise errors.InstrumentLogEntryDoesNotExistError: when no instrument log
        entry with the given ID exists
    :raise errors.InstrumentLogCategoryDoesNotExistError: when no log category
        with the given ID exists for this instrument
    """
    # ensure that the instrument exists
    instrument_log_entry = instrument_log_entries.InstrumentLogEntry.query.filter_by(id=log_entry_id).first()
    if instrument_log_entry is None:
        raise errors.InstrumentLogEntryDoesNotExistError()
    instrument_id = instrument_log_entry.instrument_id
    categories = []
    for category_id in set(category_ids):
        category = instrument_log_entries.InstrumentLogCategory.query.filter_by(
            id=category_id,
            instrument_id=instrument_id
        ).first()
        if category:
            categories.append(category)
        else:
            raise errors.InstrumentLogCategoryDoesNotExistError()

    next_version_id = max(
        v.version_id
        for v in instrument_log_entry.versions
    ) + 1

    instrument_log_entry_version = instrument_log_entries.InstrumentLogEntryVersion(
        log_entry_id=instrument_log_entry.id,
        version_id=next_version_id,
        content=content
    )
    db.session.add(instrument_log_entry_version)
    instrument_log_entry_version.categories = list(categories)
    db.session.commit()

    instrument = instruments.get_instrument(instrument_log_entry.instrument_id)

    for responsible_user in instrument.responsible_users:
        if responsible_user.id != instrument_log_entry.user_id:
            create_notification_for_an_edited_instrument_log_entry(
                user_id=responsible_user.id,
                instrument_log_entry_id=instrument_log_entry.id,
                version_id=next_version_id
            )

    return InstrumentLogEntry.from_database(instrument_log_entry)


def get_instrument_log_entry(instrument_log_entry_id: int) -> InstrumentLogEntry:
    """
    Return an individual instrument log entry.

    :param instrument_log_entry_id: the ID of an existing instrument log entry
    :raise errors.InstrumentLogEntryDoesNotExistError: when no instrument log
        entry with the given ID exists
    """
    log_entry = instrument_log_entries.InstrumentLogEntry.query.filter_by(
        id=instrument_log_entry_id
    ).first()
    if log_entry is None:
        raise errors.InstrumentLogEntryDoesNotExistError()
    return InstrumentLogEntry.from_database(log_entry)


def get_instrument_log_entries(instrument_id: int) -> typing.List[InstrumentLogEntry]:
    """
    Returns a list of log entries for an instrument.

    :param instrument_id: the ID of an existing instrument
    :return: the list of log entries, sorted from oldest to newest
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    log_entries = instrument_log_entries.InstrumentLogEntry.query.filter_by(
        instrument_id=instrument_id
    ).all()
    log_entries.sort(key=lambda e: e.versions[0].utc_datetime)
    if not log_entries:
        # ensure that the instrument exists
        instruments.get_instrument(instrument_id)
    return [
        InstrumentLogEntry.from_database(log_entry)
        for log_entry in log_entries
    ]


def create_instrument_log_file_attachment(
        instrument_log_entry_id: int,
        file_name: str,
        content: bytes
) -> None:
    """
    Create a file attachment to a instrument log entry.

    :param instrument_log_entry_id: the ID of an existing instrument log entry
    :param file_name: the original file name
    :param content: the file content
    :raise errors.InstrumentLogEntryDoesNotExistError: when no instrument log
        entry with the given ID exists
    """
    # ensure that the instrument log entry exists
    log_entry = get_instrument_log_entry(instrument_log_entry_id)
    attachment = instrument_log_entries.InstrumentLogFileAttachment(
        log_entry_id=log_entry.id,
        file_name=file_name,
        content=content
    )
    db.session.add(attachment)
    db.session.commit()


def get_instrument_log_file_attachments(
        instrument_log_entry_id: int
) -> typing.List[InstrumentLogFileAttachment]:
    """
    Return a list of all file attachments for an instrument log entry.

    :param instrument_log_entry_id: the ID of an existing instrument log entry
    :return: the list of file attachments
    :raise errors.InstrumentLogEntryDoesNotExistError: when no instrument log
        entry with the given ID exists
    """
    # ensure that the instrument log entry exists
    log_entry = get_instrument_log_entry(instrument_log_entry_id)
    attachments = instrument_log_entries.InstrumentLogFileAttachment.query.filter_by(
        log_entry_id=log_entry.id
    ).order_by(instrument_log_entries.InstrumentLogFileAttachment.id).all()
    return [
        InstrumentLogFileAttachment.from_database(attachment)
        for attachment in attachments
    ]


def get_instrument_log_file_attachment(
        instrument_log_file_attachment_id: int
) -> InstrumentLogFileAttachment:
    """
    Return an instrument log entry file attachment.

    :param instrument_log_file_attachment_id: the ID of an existing instrument
        log file attachment
    :return: the file attachment with the given ID
    :raise errors.InstrumentLogFileAttachmentDoesNotExistError: when no
        instrument log entry file attachment with the given ID exists
    """
    attachment = instrument_log_entries.InstrumentLogFileAttachment.query.filter_by(
        id=instrument_log_file_attachment_id
    ).first()
    if attachment is None:
        raise errors.InstrumentLogFileAttachmentDoesNotExistError()
    return InstrumentLogFileAttachment.from_database(attachment)


def create_instrument_log_object_attachment(
        instrument_log_entry_id: int,
        object_id: int
) -> None:
    """
    Create an object attachment to a instrument log entry.

    :param instrument_log_entry_id: the ID of an existing instrument log entry
    :param object_id: the ID of an existing object
    :raise errors.InstrumentLogEntryDoesNotExistError: when no instrument log
        entry with the given ID exists
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    # ensure that the instrument log entry exists
    log_entry = get_instrument_log_entry(instrument_log_entry_id)
    # ensure the object exists
    objects.get_object(object_id)
    attachment = instrument_log_entries.InstrumentLogObjectAttachment(
        log_entry_id=log_entry.id,
        object_id=object_id
    )
    db.session.add(attachment)
    db.session.commit()


def get_instrument_log_object_attachments(
        instrument_log_entry_id: int
) -> typing.List[InstrumentLogObjectAttachment]:
    """
    Return a list of all object attachments for an instrument log entry.

    :param instrument_log_entry_id: the ID of an existing instrument log entry
    :return: the list of object attachments
    :raise errors.InstrumentLogEntryDoesNotExistError: when no instrument log
        entry with the given ID exists
    """
    # ensure that the instrument log entry exists
    log_entry = get_instrument_log_entry(instrument_log_entry_id)
    attachments = instrument_log_entries.InstrumentLogObjectAttachment.query.filter_by(
        log_entry_id=log_entry.id
    ).order_by(instrument_log_entries.InstrumentLogObjectAttachment.id).all()
    return [
        InstrumentLogObjectAttachment.from_database(attachment)
        for attachment in attachments
    ]


def get_instrument_log_object_attachment(
        instrument_log_object_attachment_id: int
) -> InstrumentLogObjectAttachment:
    """
    Return an instrument log entry object attachment.

    :param instrument_log_object_attachment_id: the ID of an existing
        instrument log object attachment
    :return: the object attachment with the given ID
    :raise errors.InstrumentLogObjectAttachmentDoesNotExistError: when no
        instrument log entry object attachment with the given ID exists
    """
    attachment = instrument_log_entries.InstrumentLogObjectAttachment.query.filter_by(
        id=instrument_log_object_attachment_id
    ).first()
    if attachment is None:
        raise errors.InstrumentLogObjectAttachmentDoesNotExistError()
    return InstrumentLogObjectAttachment.from_database(attachment)


def hide_instrument_log_object_attachment(
        instrument_log_object_attachment_id: int,
        set_hidden: bool = True
) -> None:
    """
    Hide an instrument log entry object attachment.

    :param instrument_log_object_attachment_id: the ID of an existing
        instrument log object attachment
    :param set_hidden: whether or not the attachment should be hidden
    :raise errors.InstrumentLogObjectAttachmentDoesNotExistError: when no
        instrument log entry object attachment with the given ID exists
    """
    attachment = instrument_log_entries.InstrumentLogObjectAttachment.query.filter_by(
        id=instrument_log_object_attachment_id
    ).first()
    if attachment is None:
        raise errors.InstrumentLogObjectAttachmentDoesNotExistError()
    attachment.is_hidden = set_hidden
    db.session.add(attachment)
    db.session.commit()


def hide_instrument_log_file_attachment(
        instrument_log_file_attachment_id: int
) -> None:
    """
    Hide an instrument log entry file attachment.

    :param instrument_log_file_attachment_id: the ID of an existing
        instrument log file attachment
    :raise errors.InstrumentLogFileAttachmentDoesNotExistError: when no
        instrument log entry file attachment with the given ID exists
    """
    attachment = instrument_log_entries.InstrumentLogFileAttachment.query.filter_by(
        id=instrument_log_file_attachment_id
    ).first()
    if attachment is None:
        raise errors.InstrumentLogFileAttachmentDoesNotExistError()
    attachment.is_hidden = True
    db.session.add(attachment)
    db.session.commit()



# coding: utf-8
"""
Add confirmed column to object location assignments table.
"""

import os

MIGRATION_INDEX = 11
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    column_names = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'object_location_assignments'
    """).fetchall()
    if ('confirmed',) in column_names:
        return False

    # Perform migration
    db.session.execute("""
        ALTER TABLE object_location_assignments
        ADD confirmed BOOLEAN
    """)
    db.session.execute("""
        UPDATE object_location_assignments
        SET confirmed = FALSE
        WHERE confirmed IS NULL
    """)
    db.session.execute("""
        ALTER TABLE object_location_assignments
        ALTER COLUMN confirmed SET NOT NULL
    """)
    db.session.execute("""
        ALTER TABLE object_location_assignments
        ALTER COLUMN confirmed SET DEFAULT FALSE
    """)
    return True



# coding: utf-8
"""
Dublin Core metadata in RDF/XML format
"""

import typing

import flask
import jinja2

from . import objects, object_log, users

RDF_TEMPLATE = jinja2.Template("""<?xml version="1.0"?>
<rdf:RDF
xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
xmlns:dcterms="http://purl.org/dc/terms/"
xmlns:foaf="http://xmlns.com/foaf/0.1/"
>
  <rdf:Description xml:lang="en" rdf:about="{{ object_url }}">
    <dcterms:identifier>{{ object_url }}</dcterms:identifier>
    <dcterms:title>{{object_name}}</dcterms:title>
    <dcterms:created>{{ created.isoformat() }}</dcterms:created>
    <dcterms:modified>{{ modified.isoformat() }}</dcterms:modified>

{% for user in creators %}
    <dcterms:creator>
      <foaf:Person rdf:about="{{ user['url'] }}">
        <foaf:name>{{ user['name'] }}</foaf:name>
      </foaf:Person>
    </dcterms:creator>
{% endfor %}

{% for user in contributors %}
    <dcterms:contributor>
      <foaf:Person rdf:about="{{ user['url'] }}">
        <foaf:name>{{ user['name'] }}</foaf:name>
      </foaf:Person>
    </dcterms:contributor>
{% endfor %}

{% for version_url in version_urls %}
    <dcterms:hasVersion rdf:resource="{{ version_url }}" />
{% endfor %}

{% if current_object_url %}
    <dcterms:isVersionOf rdf:resource="{{ current_object_url }}" />
{% endif %}
  </rdf:Description>
</rdf:RDF>
""")


def generate_rdf(user_id: int, object_id: int, version_id: typing.Optional[int] = None) -> str:
    """
    Generate an XML file for an object.

    The file will contain an RDF description for an object using Dublin Core metadata.

    :param user_id: the ID of an existing user
    :param object_id: the ID of the existing object
    :param version_id: the ID of the object's existing version
    :return:
    :raise errors.UserDoesNotExistError: when no user with the given ID exists
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.ObjectVersionDoesNotExistError: when an object with the
        given object ID exists, but does not have a version with the given
        version ID
    """

    object = objects.get_object(object_id, version_id)

    log_entries = object_log.get_object_log_entries(object.id, user_id)

    creation_datetime = object.utc_datetime
    modification_datetime = object.utc_datetime

    creator_ids = []
    contributor_ids = []
    for entry in reversed(log_entries):
        if entry.utc_datetime > modification_datetime:
            modification_datetime = entry.utc_datetime
        if entry.type in {
            object_log.ObjectLogEntryType.CREATE_BATCH,
            object_log.ObjectLogEntryType.CREATE_OBJECT,
            object_log.ObjectLogEntryType.EDIT_OBJECT,
            object_log.ObjectLogEntryType.RESTORE_OBJECT_VERSION,
        }:
            if entry.user_id not in creator_ids:
                creator_ids.append(entry.user_id)
            if entry.type in {
                object_log.ObjectLogEntryType.CREATE_OBJECT,
                object_log.ObjectLogEntryType.CREATE_BATCH
            }:
                creation_datetime = entry.utc_datetime
                if version_id == 0:
                    break

            if entry.type in {
                object_log.ObjectLogEntryType.EDIT_OBJECT,
                object_log.ObjectLogEntryType.RESTORE_OBJECT_VERSION,
            } and entry.data['version_id'] == version_id:
                break
        else:
            if entry.user_id not in contributor_ids:
                contributor_ids.append(entry.user_id)

    creators = [
        users.get_user(user_id)
        for user_id in creator_ids
    ]

    contributors = [
        users.get_user(user_id)
        for user_id in contributor_ids
        if user_id not in creator_ids
    ]

    if version_id is None:
        object_url = flask.url_for(
            'frontend.object',
            object_id=object_id,
            _external=True
        )
        version_urls = [
            flask.url_for(
                'frontend.object_version',
                object_id=object_id,
                version_id=other_version_id,
                _external=True
            )
            for other_version_id in range(object.version_id + 1)
        ]
        current_object_url = None
    else:
        object_url = flask.url_for(
            'frontend.object_version',
            object_id=object_id,
            version_id=version_id,
            _external=True
        )
        version_urls = []
        current_object_url = flask.url_for(
            'frontend.object',
            object_id=object_id,
            _external=True
        )

    object_name = object.data.get('name', {}).get('text', 'Unnamed Object')

    return RDF_TEMPLATE.render(
        object_url=object_url,
        object_name=object_name,
        created=creation_datetime,
        modified=modification_datetime,
        creators=[
            {
                'name': user.name,
                'url': flask.url_for('frontend.user_profile', user_id=user.id, _external=True)
            }
            for user in creators
        ],
        contributors=[
            {
                'name': user.name,
                'url': flask.url_for('frontend.user_profile', user_id=user.id, _external=True)
            }
            for user in contributors
        ],
        version_urls=version_urls,
        current_object_url=current_object_url
    )



# coding: utf-8
"""
RESTful API for SampleDB
"""

import json
import flask

from flask_restful import Resource

from ...api.server.authentication import multi_auth, object_permissions_required, Permissions
from ...logic.actions import get_action, get_action_type
from ...logic.object_search import generate_filter_func, wrap_filter_func
from ...logic.objects import get_object, update_object, create_object
from ...logic.object_permissions import get_objects_with_permissions
from ...logic import errors
from ... import models

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class ObjectVersion(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, version_id: int):
        try:
            object = get_object(object_id=object_id, version_id=version_id)
        except errors.ObjectVersionDoesNotExistError:
            return {
                "message": "version {} of object {} does not exist".format(version_id, object_id)
            }, 404
        return {
            'object_id': object.object_id,
            'version_id': object.version_id,
            'action_id': object.action_id,
            'schema': object.schema,
            'data': object.data
        }


class ObjectVersions(Resource):
    @object_permissions_required(Permissions.WRITE)
    def post(self, object_id: int):
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        for key in request_json:
            if key not in {'object_id', 'version_id', 'action_id', 'schema', 'data'}:
                return {
                    "message": "invalid key '{}'".format(key)
                }, 400
        object = get_object(object_id=object_id)
        action = get_action(action_id=object.action_id)
        if 'object_id' in request_json:
            if request_json['object_id'] != object.object_id:
                return {
                    "message": "object_id must be {}".format(object.object_id)
                }, 400
        if 'version_id' in request_json:
            if request_json['version_id'] != object.version_id + 1:
                return {
                    "message": "version_id must be {}".format(object.version_id + 1)
                }, 400
        if 'action_id' in request_json:
            if request_json['action_id'] != object.action_id:
                return {
                    "message": "action_id must be {}".format(object.action_id)
                }, 400
        if 'schema' in request_json:
            if request_json['schema'] != object.schema and request_json['schema'] != action.schema:
                return {
                    "message": "schema must be either:\n{}\nor:\n{}".format(json.dumps(object.schema, indent=4), json.dumps(action.schema, indent=4))
                }, 400
            schema = request_json['schema']
        else:
            schema = object.schema
        if 'data' not in request_json:
            return {
                "message": "data must be set"
            }, 400
        data = request_json['data']
        try:
            update_object(
                object_id=object.object_id,
                data=data,
                user_id=flask.g.user.id,
                schema=schema
            )
        except errors.ValidationError as e:
            messages = e.message.splitlines()
            return {
                "message": "validation failed:\n - {}".format('\n - '.join(messages))
            }, 400
        except Exception:
            return {
                "message": "failed to update object"
            }, 400
        object_version_url = flask.url_for(
            'api.object_version',
            object_id=object.object_id,
            version_id=object.version_id + 1
        )
        return flask.redirect(object_version_url, code=201)


class Object(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int):
        object = get_object(object_id=object_id)
        object_version_url = flask.url_for(
            'api.object_version',
            object_id=object.object_id,
            version_id=object.version_id
        )
        return flask.redirect(object_version_url, code=302)


class Objects(Resource):
    @multi_auth.login_required
    def get(self):
        action_id = flask.request.args.get('action_id', '')
        if action_id:
            try:
                action_id = int(action_id)
            except ValueError:
                return {
                    'message': 'Unable to parse action_id'
                }, 400
            try:
                get_action(action_id)
            except errors.ActionDoesNotExistError:
                return {
                    'message': 'No action with the given action_id exists.'
                }, 400
        else:
            action_id = None
        action_type_id = flask.request.args.get('action_type_id', flask.request.args.get('action_type', None))
        if action_type_id is not None:
            try:
                action_type_id = int(action_type_id)
            except ValueError:
                # ensure old links still function
                action_type_id = {
                    'sample': models.ActionType.SAMPLE_CREATION,
                    'measurement': models.ActionType.MEASUREMENT,
                    'simulation': models.ActionType.SIMULATION
                }.get(action_type_id, None)
            else:
                try:
                    get_action_type(action_type_id)
                except errors.ActionTypeDoesNotExistError:
                    action_type_id = None
            if action_type_id is None:
                return {
                    'message': 'No matching action type exists.'
                }, 400
        else:
            action_type_id = None
        project_id = None
        limit = flask.request.args.get('limit')
        if limit is not None:
            try:
                limit = int(limit)
            except ValueError:
                limit = None
        if limit is not None and not 1 <= limit < 1e15:
            limit = None
        offset = flask.request.args.get('offset')
        if offset is not None:
            try:
                offset = int(offset)
            except ValueError:
                offset = None
        if offset is not None and not 0 <= offset < 1e15:
            offset = None
        name_only = bool(flask.request.args.get('name_only'))
        query_string = flask.request.args.get('q', '')
        if query_string:
            try:
                filter_func, search_tree, use_advanced_search = generate_filter_func(query_string, True)
            except Exception:
                # TODO: ensure that advanced search does not cause exceptions
                def filter_func(data, search_notes):
                    """ Return all objects"""
                    search_notes.append(('error', "Unable to parse search expression", 0, len(query_string)))
                    return False
            filter_func, search_notes = wrap_filter_func(filter_func)
        else:
            search_notes = []

            def filter_func(data):
                return True
        try:
            objects = get_objects_with_permissions(
                user_id=flask.g.user.id,
                permissions=Permissions.READ,
                filter_func=filter_func,
                action_id=action_id,
                action_type_id=action_type_id,
                project_id=project_id,
                limit=limit,
                offset=offset,
                name_only=name_only
            )
        except Exception as e:
            search_notes.append(('error', "Error during search: {}".format(e), 0, 0))
            objects = []
        if any(search_note[0] == 'error' for search_note in search_notes):
            return {
                'message': '\n'.join(
                    'Error: ' + search_note[1]
                    for search_note in search_notes
                    if search_note[0] == 'error'
                )
            }, 400
        else:
            return [
                {
                    'object_id': object.object_id,
                    'version_id': object.version_id,
                    'action_id': object.action_id,
                    'schema': object.schema,
                    'data': object.data
                }
                for object in objects
            ], 200

    @multi_auth.login_required
    def post(self):
        if flask.g.user.is_readonly:
            return {
                'message': 'user has been marked as read only'
            }, 400
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        for key in request_json:
            if key not in {'object_id', 'version_id', 'action_id', 'schema', 'data'}:
                return {
                    "message": "invalid key '{}'".format(key)
                }, 400
        if 'object_id' in request_json:
            return {
                "message": "object_id cannot be set"
            }, 400
        if 'version_id' in request_json:
            if request_json['version_id'] != 0:
                return {
                    "message": "version_id must be 0"
                }, 400
        if 'action_id' in request_json:
            if not isinstance(request_json['action_id'], int):
                return {
                    "message": "action_id must be an integer"
                }, 400
            action_id = request_json['action_id']
            try:
                action = get_action(action_id=action_id)
            except errors.ActionDoesNotExistError:
                return {
                    "message": "action {} does not exist".format(action_id)
                }, 400
        else:
            return {
                "message": "action_id must be set"
            }, 400
        if 'schema' in request_json:
            if request_json['schema'] != action.schema:
                return {
                    "message": "schema must be:\n{}".format(json.dumps(action.schema, indent=4))
                }, 400
        schema = action.schema
        if 'data' not in request_json:
            return {
                "message": "data must be set"
            }, 400
        data = request_json['data']
        try:
            object = create_object(
                action_id=action_id,
                data=data,
                user_id=flask.g.user.id,
                schema=schema
            )
        except errors.ValidationError as e:
            messages = e.message.splitlines()
            return {
                "message": "validation failed:\n - {}".format('\n - '.join(messages))
            }, 400
        except Exception:
            return {
                "message": "failed to create object"
            }, 400
        object_version_url = flask.url_for(
            'api.object_version',
            object_id=object.object_id,
            version_id=object.version_id
        )
        return flask.redirect(object_version_url, code=201)



# coding: utf-8
"""
Script for updating the users responsible for an instrument in SampleDB.

Usage: python -m sampledb update_instrument_responsible_users <instrument_id> <instrument_reponsible_user_ids...>
"""

import sys
from .. import create_app
from ..logic.instruments import get_instrument, add_instrument_responsible_user, remove_instrument_responsible_user
from ..logic.users import get_user
from ..logic.errors import UserDoesNotExistError, InstrumentDoesNotExistError


def main(arguments):
    if len(arguments) == 0:
        print(__doc__)
        exit(1)
    instrument_id = arguments[0]
    instrument_responsible_user_ids = arguments[1:]
    try:
        instrument_id = int(instrument_id)
    except ValueError:
        print("Error: instrument_id must be an integer", file=sys.stderr)
        exit(1)
    for i, user_id in enumerate(instrument_responsible_user_ids):
        try:
            user_id = int(user_id)
        except ValueError:
            print("Error: instrument_reponsible_user_ids must be integer", file=sys.stderr)
            exit(1)
        instrument_responsible_user_ids[i] = user_id
    app = create_app()
    with app.app_context():
        try:
            instrument = get_instrument(instrument_id)
        except InstrumentDoesNotExistError:
            print('Error: no instrument with this id exists', file=sys.stderr)
            exit(1)
        for user_id in instrument_responsible_user_ids:
            try:
                get_user(user_id)
            except UserDoesNotExistError:
                print('Error: no user with the id #{} exists'.format(user_id), file=sys.stderr)
                exit(1)
        previous_instrument_responsible_user_ids = [user.id for user in instrument.responsible_users]
        for user_id in instrument_responsible_user_ids:
            if user_id not in previous_instrument_responsible_user_ids:
                add_instrument_responsible_user(instrument.id, user_id)
        for user_id in previous_instrument_responsible_user_ids:
            if user_id not in instrument_responsible_user_ids:
                remove_instrument_responsible_user(instrument.id, user_id)
        print("Success: the instrument responsible users have been updated in SampleDB")



# coding: utf-8
"""

"""

import collections
import datetime
import sqlalchemy as db
import sqlalchemy.dialects.postgresql as postgresql

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class VersionedJSONSerializableObjectTables(object):
    """
    A class for storing JSON-serializable objects without deletes and with versioned updates.

    The class uses two tables, one for the current objects and one for previous versions of these objects. Each
    object consists of an object ID, a version ID, the actual data, a JSON schema for the data and information about
    this current version (the ID of this version's author and a datetime of when the object was created or updated).

    These two tables will be created when the instance of this class is created, if bind is provided. Otherwise you
    can use the metadata attribute to access the SQLAlchemy MetaData object associated with these tables. You can then
    call metadata.create_all(bind) to create the tables.

    You should **not** interact with the tables yourself. Instead, use the functions provided by this class.

    For information on JSON schemas, see http://json-schema.org/.
    """

    class VersionedJSONSerializableObject(collections.namedtuple(
            'VersionedJSONSerializableObject',
            [
                'object_id',
                'version_id',
                'action_id',
                'data',
                'schema',
                'user_id',
                'utc_datetime'
            ]
    )):

        @property
        def id(self) -> int:
            return self.object_id

    def __init__(self, table_name_prefix, bind=None, object_type=VersionedJSONSerializableObject, user_id_column=None, action_id_column=None, action_schema_column=None, metadata=None, create_object_callbacks=None, data_validator=None, schema_validator=None):
        """
        Creates new instance for storing versioned, JSON-serializable objects using two tables.

        :param table_name_prefix: the prefix used for naming the two used tables
        :param bind: the SQLAlchemy engine used for creating the tables and for future connections
        :param object_type: the type used for returning objects
        :param user_id_column: a SQLAlchemy column object for use as foreign key for the user ID (optional)
        :param action_id_column: a SQLAlchemy column object for use as foreign key for the action ID (optional)
        :param metdata: an SQLAlchemy MetaData object used for creating the two tables (optional)
        :param create_object_callbacks: a list of callables which will be called when an object is created (optional)
        :param data_validator: a data validator function (given the data and the schema) (optional)
        :param schema_validator: a schema validator function (given the schema) (optional)
        """
        if metadata is None:
            metadata = db.MetaData()
        self.metadata = metadata
        self._current_table = db.Table(
            table_name_prefix + '_current',
            self.metadata,
            db.Column('object_id', db.Integer, nullable=False, primary_key=True, autoincrement=True),
            db.Column('version_id', db.Integer, nullable=False, default=0),
            db.Column('action_id', db.Integer, nullable=False),
            db.Column('data', postgresql.JSONB, nullable=False),
            db.Column('schema', postgresql.JSONB, nullable=False),
            db.Column('user_id', db.Integer, nullable=False),
            db.Column('utc_datetime', db.DateTime, nullable=False)
        )
        self._previous_table = db.Table(
            table_name_prefix + '_previous',
            self.metadata,
            db.Column('object_id', db.Integer, nullable=False),
            db.Column('version_id', db.Integer, nullable=False),
            db.Column('action_id', db.Integer, nullable=False),
            db.Column('data', postgresql.JSONB, nullable=False),
            db.Column('schema', postgresql.JSONB, nullable=False),
            db.Column('user_id', db.Integer, nullable=False),
            db.Column('utc_datetime', db.DateTime, nullable=False),
            db.PrimaryKeyConstraint('object_id', 'version_id')
        )
        if user_id_column is not None:
            self._current_table.append_constraint(db.ForeignKeyConstraint(['user_id'], [user_id_column]))
            self._previous_table.append_constraint(db.ForeignKeyConstraint(['user_id'], [user_id_column]))
        if action_id_column is not None:
            self._current_table.append_constraint(db.ForeignKeyConstraint(['action_id'], [action_id_column]))
            self._previous_table.append_constraint(db.ForeignKeyConstraint(['action_id'], [action_id_column]))
        elif action_schema_column is not None:
            raise ValueError("Using action_schema_column requires using the action_id_column as well!")
        self._action_id_column = action_id_column
        self._action_schema_column = action_schema_column
        self.object_type = object_type
        self.object_id_column = self._current_table.c.object_id
        self.bind = bind
        if self.bind is not None:
            self.metadata.create_all(self.bind)
        self._data_validator = data_validator
        self._schema_validator = schema_validator

    def create_object(self, data, schema, user_id, action_id, utc_datetime=None, connection=None):
        """
        Creates an object in the table for current objects. This object will always have version_id 0.

        :param data: a JSON serializable object containing the object data
        :param schema: a JSON schema describing data (may be None if action's schema is to be used)
        :param user_id: the ID of the user who created the object
        :param action_id: the ID of the action which was used to create the object
        :param utc_datetime: the datetime (in UTC) when the object was created (optional, defaults to utcnow())
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the newly created object as object_type
        """
        if connection is None:
            connection = self.bind.connect()
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        if schema is None and self._action_schema_column is not None:
            action = connection.execute(
                db
                .select([
                    self._action_schema_column
                ])
                .where(self._action_id_column == action_id)
            ).fetchone()
            if action is None:
                raise ValueError('Action with id {} not found'.format(action_id))
            schema = action[0]
        if self._schema_validator:
            self._schema_validator(schema)
        if self._data_validator:
            self._data_validator(data, schema)
        version_id = 0
        object_id = connection.execute(
            self._current_table
            .insert()
            .values(
                version_id=version_id,
                action_id=action_id,
                data=data,
                schema=schema,
                user_id=user_id,
                utc_datetime=utc_datetime
            )
            .returning(
                self._current_table.c.object_id
            )
        ).scalar()
        obj = self.object_type(
            object_id=object_id,
            version_id=version_id,
            action_id=action_id,
            data=data,
            schema=schema,
            user_id=user_id,
            utc_datetime=utc_datetime
        )
        return obj

    def update_object(self, object_id, data, schema, user_id, utc_datetime=None, connection=None):
        """
        Updates an existing object using the given data, user id and datetime.

        This step first copies the existing object in the table of previous object versions and then modifies the
        object's version id, incrementing it by one, and data, replacing it by the given data.

        :param object_id: the ID of the existing object
        :param data: a JSON serializable object containing the updated object data
        :param schema: a JSON schema describing data (optional, defaults to the current object schema)
        :param user_id: the ID of the user who updated the object
        :param utc_datetime: the datetime (in UTC) when the object was updated (optional, defaults to utcnow())
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the updated object as object_type or None, if the object does not exist
        """
        if connection is None:
            connection = self.bind.connect()
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        if schema is None:
            schema = connection.execute(
                db.select([self._current_table.c.schema]).where(self._current_table.c.object_id == object_id)
            ).fetchone()
            if schema is None:
                return None
            schema = schema[0]
        if self._schema_validator:
            self._schema_validator(schema)
        if self._data_validator:
            self._data_validator(data, schema)
        with connection.begin() as transaction:
            # Copy current version to previous versions
            if connection.execute(
                self._previous_table
                .insert()
                .from_select(
                    ['object_id', 'version_id', 'action_id', 'data', 'schema', 'user_id', 'utc_datetime'],
                    self._current_table
                    .select()
                    .where(self._current_table.c.object_id == db.bindparam('oid'))
                ),
                [{'oid': object_id}]
            ).rowcount != 1:
                return None
            # Update current version to new version
            connection.execute(
                self._current_table
                .update()
                .where(self._current_table.c.object_id == db.bindparam('oid'))
                .values(
                    version_id=self._current_table.c.version_id + 1,
                    data=data,
                    schema=schema,
                    user_id=user_id,
                    utc_datetime=utc_datetime
                ),
                [{'oid': object_id}]
            )
            transaction.commit()
        return self.get_current_object(object_id, connection=connection)

    def get_current_object(self, object_id, connection=None):
        """
        Queries and returns an object by its ID.

        :param object_id: the ID of the existing object
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the object as object_type or None if the object does not exist
        """
        if connection is None:
            connection = self.bind.connect()
        current_object = connection.execute(
            db
            .select([
                self._current_table.c.object_id,
                self._current_table.c.version_id,
                self._current_table.c.action_id,
                self._current_table.c.data,
                self._current_table.c.schema,
                self._current_table.c.user_id,
                self._current_table.c.utc_datetime
            ])
            .where(self._current_table.c.object_id == object_id)
        ).fetchone()
        if current_object is None:
            return None
        return self.object_type(*current_object)

    def get_current_objects(self, filter_func=lambda data: True, action_table=None, action_filter=None, connection=None, table=None, parameters=None, sorting_func=None, limit=None, offset=None, num_objects_found=None):
        """
        Queries and returns all objects matching a given filter.

        :param filter_func: a lambda that may return an SQLAlchemy filter when given a table
        :param action_filter: a SQLAlchemy comparator, used to query only objects created by specific actions
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :param table: a custom SQLAlchemy table-like object to use as base for the query (optional)
        :param parameters: query parameters for the custom select statement (optional)
        :return: a list of objects as object_type
        """
        if connection is None:
            connection = self.bind.connect()

        if parameters is None:
            parameters = {}

        if table is None:
            table = self._current_table

        select_statement = db.select([
            table.c.object_id,
            table.c.version_id,
            table.c.action_id,
            table.c.data,
            table.c.schema,
            table.c.user_id,
            table.c.utc_datetime,
            db.sql.expression.text('COUNT(*) OVER()')
        ])

        selectable = table

        if sorting_func is not None and getattr(sorting_func, 'require_original_columns', False):
            selectable = selectable.outerjoin(
                self._previous_table,
                db.and_(table.c.object_id == self._previous_table.c.object_id, self._previous_table.c.version_id == 0),
                full=False
            )

        if action_table is not None and action_filter is not None:
            assert self._action_id_column is not None
            assert action_table is not None
            assert action_filter is not None
            selectable = selectable.join(
                action_table,
                db.and_(table.c.action_id == self._action_id_column, action_filter)
            )

        select_statement = select_statement.select_from(selectable)

        if sorting_func is None:
            def sorting_func(current_columns, original_columns):
                return db.sql.desc(current_columns.object_id)

        select_statement = select_statement.where(filter_func(table.c.data))
        select_statement = select_statement.order_by(sorting_func(table.c, self._previous_table.c))

        if limit is not None:
            select_statement = select_statement.limit(limit)

        if offset is not None:
            select_statement = select_statement.offset(offset)

        objects = connection.execute(
            select_statement,
            **parameters
        ).fetchall()
        if num_objects_found is not None:
            num_objects_found.clear()
            if objects:
                num_objects_found.append(objects[0][-1])
            else:
                num_objects_found.append(0)
        return [self.object_type(*obj[:-1]) for obj in objects]

    def get_object_versions(self, object_id, connection=None):
        """
        Queries and returns all versions of an object with a given ID, sorted ascendingly by the version ID, from first
        version to current version.

        :param object_id: the ID of the existing object
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: a list of objects as object_type
        """
        if connection is None:
            connection = self.bind.connect()
        current_object = self.get_current_object(object_id, connection=connection)
        if current_object is None:
            return []
        previous_objects = connection.execute(
            db
            .select([
                self._previous_table.c.object_id,
                self._previous_table.c.version_id,
                self._previous_table.c.action_id,
                self._previous_table.c.data,
                self._previous_table.c.schema,
                self._previous_table.c.user_id,
                self._previous_table.c.utc_datetime
            ])
            .where(self._previous_table.c.object_id == object_id)
            .order_by(db.asc(self._previous_table.c.version_id))
        ).fetchall()
        objects = []
        for obj in previous_objects:
            objects.append(self.object_type(*obj))
        objects.append(current_object)
        return objects

    def get_object_version(self, object_id, version_id, connection=None):
        """
        Queries and returns an individual version of an object with a given object and version ID.

        :param object_id: the ID of the existing object
        :param version_id: the ID of the object's existing version
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the object as object_type or None if the object or this version of it does not exist
        """
        if connection is None:
            connection = self.bind.connect()
        previous_objects = connection.execute(
            db
            .select([
                self._previous_table.c.object_id,
                self._previous_table.c.version_id,
                self._previous_table.c.action_id,
                self._previous_table.c.data,
                self._previous_table.c.schema,
                self._previous_table.c.user_id,
                self._previous_table.c.utc_datetime
            ])
            .where(db.and_(
                self._previous_table.c.object_id == object_id,
                self._previous_table.c.version_id == version_id
            ))
        ).fetchall()
        if previous_objects:
            return self.object_type(*previous_objects[0])
        current_object = self.get_current_object(object_id, connection=connection)
        if current_object is not None and current_object.version_id == version_id:
            return current_object
        return None



# coding: utf-8
"""
Script for sending an announcement notification to all users.

Usage: python -m sampledb send_announcement <text_file_name> <html_file_name>
"""

from .. import create_app
from ..logic.notifications import create_announcement_notification_for_all_users


def main(arguments):
    if len(arguments) != 2:
        print(__doc__)
        exit(1)
    text_file_name, html_file_name = arguments
    with open(text_file_name, 'r') as text_file:
        text = text_file.read()
    with open(html_file_name, 'r') as html_file:
        html = html_file.read()
    app = create_app()
    with app.app_context():
        create_announcement_notification_for_all_users(text, html)



# coding: utf-8
"""

"""

import base64
from io import BytesIO
import os
from urllib.parse import quote_plus

import bleach
import flask
import flask_login
import markdown
import qrcode
import qrcode.image.svg

from ..logic.errors import UserIsReadonlyError
from ..logic.units import prettify_units
from ..logic.notifications import get_num_notifications


def jinja_filter(func):
    global _jinja_filters
    _jinja_filters[func.__name__] = func
    return func


_jinja_filters = {}
jinja_filter.filters = _jinja_filters


qrcode_cache = {}


def generate_qrcode(url: str, should_cache: bool = True) -> str:
    """
    Generate a QR code (as data URI) to a given URL.

    :param url: the url the QR code should reference
    :param should_cache: whether or not the QR code should be cached
    :return: a data URI to a base64 encoded SVG image
    """
    if should_cache and url in qrcode_cache:
        return qrcode_cache[url]
    image = qrcode.make(url, image_factory=qrcode.image.svg.SvgPathFillImage)
    image_stream = BytesIO()
    image.save(image_stream)
    image_stream.seek(0)
    qrcode_url = 'data:image/svg+xml;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')
    if should_cache:
        qrcode_cache[url] = qrcode_url
    return qrcode_url


def markdown_to_safe_html(markdown_text):
    return markdown.markdown(
        bleach.clean(markdown_text),
        extensions=['tables']
    )


def has_preview(file):
    if file.storage != 'local':
        return False
    file_name = file.original_file_name
    file_extension = os.path.splitext(file_name)[1]
    return file_extension in flask.current_app.config.get('MIME_TYPES', {})


def file_name_is_image(file_name):
    file_extension = os.path.splitext(file_name)[1]
    return flask.current_app.config.get('MIME_TYPES', {}).get(file_extension, '').startswith('image/')


def is_image(file):
    if file.storage != 'local':
        return False
    return file_name_is_image(file.original_file_name)


def attachment_is_image(file_attachment):
    return file_name_is_image(file_attachment.file_name)


def get_num_unread_notifications(user):
    return get_num_notifications(user.id, unread_only=True)


def check_current_user_is_not_readonly():
    if flask_login.current_user.is_readonly:
        raise UserIsReadonlyError()


_jinja_filters['prettify_units'] = prettify_units
_jinja_filters['has_preview'] = has_preview
_jinja_filters['is_image'] = is_image
_jinja_filters['attachment_is_image'] = attachment_is_image
_jinja_filters['get_num_unread_notifications'] = get_num_unread_notifications
_jinja_filters['urlencode'] = quote_plus



import itsdangerous

MAX_AGE = 60 * 60 * 24 * 2


def generate_token(data, salt, secret_key):
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key)
    return serializer.dumps(data, salt=salt)


def verify_token(token, salt, secret_key, expiration=MAX_AGE):
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key)
    try:
        return serializer.loads(
            token,
            salt=salt,
            max_age=expiration
        )
    except itsdangerous.BadData:
        return None



# coding: utf-8
"""

"""

import typing

from .. import db

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

instrument_user_association_table = db.Table(
    'association',
    db.metadata,
    db.Column('instrument_id', db.Integer, db.ForeignKey('instruments.id')),
    db.Column('user_id', db.Integer, db.ForeignKey('users.id'))
)


class Instrument(db.Model):
    __tablename__ = 'instruments'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, nullable=False, unique=True)
    description = db.Column(db.String, nullable=False, default='')
    responsible_users = db.relationship("User", secondary=instrument_user_association_table, order_by="User.name")
    description_as_html = db.Column(db.String, nullable=True, default=None)
    users_can_create_log_entries = db.Column(db.Boolean, nullable=False, default=False)
    users_can_view_log_entries = db.Column(db.Boolean, nullable=False, default=False)
    notes = db.Column(db.String, nullable=False, default='')
    notes_as_html = db.Column(db.String, nullable=True, default=None)
    create_log_entry_default = db.Column(db.Boolean, nullable=False, default=False)
    is_hidden = db.Column(db.Boolean, nullable=False, default=False)

    def __init__(
            self,
            name: str,
            description: str = '',
            description_as_html: typing.Optional[str] = None,
            users_can_create_log_entries: bool = False,
            users_can_view_log_entries: bool = False,
            notes: str = '',
            notes_as_html: typing.Optional[str] = None,
            create_log_entry_default: bool = False,
            is_hidden: bool = False
    ):
        self.name = name
        self.description = description
        self.description_as_html = description_as_html
        self.users_can_create_log_entries = users_can_create_log_entries
        self.users_can_view_log_entries = users_can_view_log_entries
        self.notes = notes
        self.notes_as_html = notes_as_html
        self.create_log_entry_default = create_log_entry_default
        self.is_hidden = is_hidden

    def __eq__(self, other):
        return (
            self.id == other.id and
            self.name == other.name and
            self.description == other.description and
            self.description_as_html == other.description_as_html and
            self.users_can_create_log_entries == other.users_can_create_log_entries and
            self.users_can_view_log_entries == other.users_can_view_log_entries and
            self.notes == other.notes and
            self.notes_as_html == other.notes_as_html and
            self.create_log_entry_default == other.create_log_entry_default and
            self.is_hidden == other.is_hidden and
            self.responsible_users == other.responsible_users
        )

    def __repr__(self):
        return '<{0}(id={1.id}, name={1.name})>'.format(type(self).__name__, self)



# coding: utf-8
"""
Script for creating an instrument in SampleDB.

Usage: python -m sampledb create_instrument <name> <description>
"""

import sys
from .. import create_app
from ..logic.instruments import create_instrument, get_instruments


def main(arguments):
    if len(arguments) != 2:
        print(__doc__)
        exit(1)
    name, description = arguments
    app = create_app()
    with app.app_context():
        instruments = get_instruments()
        for instrument in instruments:
            if instrument.name == name:
                print('Error: an instrument with this name already exists (#{})'.format(instrument.id), file=sys.stderr)
                exit(1)
        instrument = create_instrument(name, description)
        print("Success: the instrument has been created in SampleDB (#{})".format(instrument.id))



# coding: utf-8
"""

"""
import datetime
import typing

from . import datatypes


class Literal:
    def __init__(self, input_text: str, start_position: int, value: typing.Any) -> None:
        self.input_text = input_text
        self.start_position = start_position
        self.value = value
        self.end_position = self.start_position + len(self.input_text)

    def __repr__(self) -> str:
        return '<Literal({})>'.format(self.input_text)


class Text(Literal):
    def __init__(self, input_text: str, start_position: int, value: datatypes.Text) -> None:
        super(Text, self).__init__(input_text, start_position, value)

    def __repr__(self) -> str:
        return '<Text({})>'.format(self.value.text)


class Tag(Literal):
    def __init__(self, input_text: str, start_position: int, value: str) -> None:
        super(Tag, self).__init__(input_text, start_position, value)

    def __repr__(self) -> str:
        return '<Tag(#{})>'.format(self.value)


class Boolean(Literal):
    def __init__(self, input_text: str, start_position: int, value: datatypes.Boolean) -> None:
        super(Boolean, self).__init__(input_text, start_position, value)

    def __repr__(self) -> str:
        return '<Boolean({})>'.format(self.value.value)


class Attribute(Literal):
    def __init__(self, input_text: str, start_position: int, value: typing.List[str]) -> None:
        super(Attribute, self).__init__(input_text, start_position, value)

    def __repr__(self) -> str:
        return '<Attribute({})>'.format(self.value)


class Quantity(Literal):
    def __init__(self, input_text: str, start_position: int, value: datatypes.Quantity) -> None:
        super(Quantity, self).__init__(input_text, start_position, value)

    def __repr__(self) -> str:
        return '<Quantity({})>'.format(self.value)


class Date(Literal):
    def __init__(self, input_text: str, start_position: int, value: datatypes.DateTime) -> None:
        super(Date, self).__init__(input_text, start_position, value)

    def __repr__(self) -> str:
        return '<Date({})>'.format(self.value.utc_datetime)


class Token:
    def __init__(self, input_text: str, start_position: int) -> None:
        self.input_text = input_text
        self.start_position = start_position

    def __repr__(self) -> str:
        return '<Token("{}")>'.format(self.input_text)


class Operator:
    def __init__(self, input_text: str, start_position: int, operator: str) -> None:
        self.input_text = input_text
        self.start_position = start_position
        self.operator = operator

    def __repr__(self) -> str:
        return '<Operator(operator="{}")>'.format(self.operator)


class ParseError(Exception):
    def __init__(self, message: str, start: int, end: int) -> None:
        super(ParseError, self).__init__(message)
        self.message = message
        self.start = start
        self.end = end


def split_by_texts(tokens: typing.List[typing.Union[Token]]) -> typing.List[typing.Union[Token, Text]]:
    previous_tokens = tokens
    tokens = []
    for token in previous_tokens:
        is_in_text = False
        current_token_start_position = 0
        current_token_text = ""
        for i, c in enumerate(token.input_text):
            if not is_in_text:
                if c == '"':
                    is_in_text = True
                    if current_token_text:
                        tokens.append(Token(current_token_text, current_token_start_position))
                    current_token_text = ''
                    current_token_start_position = i
                    current_token_text += c
                else:
                    current_token_text += c
            elif is_in_text:
                if c == '"':
                    is_in_text = False
                    current_token_text += c
                    tokens.append(Text(current_token_text, current_token_start_position, datatypes.Text(current_token_text[1:-1])))
                    current_token_text = ''
                    current_token_start_position = i + 1
                else:
                    current_token_text += c
        if is_in_text:
            raise ParseError("Unfinished text", current_token_start_position, len(token.input_text))
        if current_token_text:
            tokens.append(Token(current_token_text, current_token_start_position))
    return tokens


def split_by_operators(tokens: typing.List[typing.Union[Token, Text]], operators: typing.List[str]) -> typing.List[typing.Union[Token, Text, Operator]]:
    for operator in operators:
        previous_tokens = tokens
        tokens = []
        found_operator = True
        while found_operator:
            found_operator = False
            for token in previous_tokens:
                if found_operator or not isinstance(token, Token):
                    tokens.append(token)
                    continue
                start_position = token.start_position
                if ' ' + token.input_text + ' ' == operator:
                    found_operator = True
                    operator_text = operator[1:-1]
                    before_operator, after_operator = '', ''
                elif (' ' + token.input_text).startswith(operator) and not token.input_text.startswith(operator):
                    found_operator = True
                    operator_text = operator[1:]
                    before_operator, after_operator = (' ' + token.input_text).split(operator, 1)
                elif (token.input_text + ' ').endswith(operator) and not token.input_text.endswith(operator):
                    found_operator = True
                    operator_text = operator[:-1]
                    before_operator, after_operator = (token.input_text + ' ').split(operator, 1)
                elif operator in token.input_text:
                    found_operator = True
                    operator_text = operator
                    before_operator, after_operator = token.input_text.split(operator, 1)
                else:
                    tokens.append(Token(token.input_text, start_position))
                    continue
                if found_operator:
                    leading_whitespace = len(operator_text) - len(operator_text.lstrip())
                    if leading_whitespace:
                        before_operator = before_operator + operator_text[:leading_whitespace]
                        operator_text = operator_text[leading_whitespace:]

                    trailing_whitespace = len(operator_text) - len(operator_text.rstrip())
                    if trailing_whitespace:
                        after_operator = operator_text[len(operator_text) - trailing_whitespace:] + after_operator
                        operator_text = operator_text[:len(operator_text) - trailing_whitespace]

                    if before_operator.strip():
                        tokens.append(Token(before_operator, start_position))
                    start_position += len(before_operator)

                    tokens.append(Operator(operator_text, start_position, operator_text))
                    start_position += len(operator_text)

                    if after_operator.strip():
                        tokens.append(Token(after_operator, start_position))
                    continue
            previous_tokens = tokens
            tokens = []
        tokens = previous_tokens
    return tokens


def apply_parentheses(tokens: typing.List[typing.Union[Token, Text, Operator]]) -> typing.List[typing.Union[Token, Text, Operator, list]]:
    previous_tokens = tokens
    tokens = []
    tokens_stack = [tokens]
    unopened_parentheses_stack = []
    for i, token in enumerate(previous_tokens):
        if not isinstance(token, Operator):
            tokens.append(token)
            continue
        if token.operator not in '()':
            tokens.append(token)
            continue
        if token.operator == '(':
            tokens.append([])
            tokens_stack.append(tokens)
            tokens = tokens[-1]
            unopened_parentheses_stack.append(i)
            # tokens.append(token)
            continue
        if token.operator == ')':
            if len(tokens_stack) < 2:
                raise ParseError("Unmatched closing parenthesis", token.start_position, token.start_position + len(token.input_text))
            # tokens.append(token)
            tokens_stack.pop()
            tokens = tokens_stack[-1]
            unopened_parentheses_stack.pop()
            continue
    if unopened_parentheses_stack:
        raise ParseError("Unmatched opening parenthesis", unopened_parentheses_stack[0], unopened_parentheses_stack[0] + 1)
    return tokens


def apply_binary_operator(tokens: typing.List[typing.Union[Token, Text, Operator, list]], operator: str):
    previous_tokens = tokens
    tokens = []
    skip_next_token = False
    for i, token in enumerate(previous_tokens):
        if skip_next_token:
            skip_next_token = False
            continue
        if isinstance(token, list):
            tokens.append(apply_binary_operator(token, operator))
            continue
        if not isinstance(token, Operator):
            tokens.append(token)
            continue
        if token.operator != operator:
            tokens.append(token)
            continue
        if token.operator == operator:
            if not tokens:
                raise ParseError("Binary operator without left operand", token.start_position, token.start_position + len(token.input_text))
            left_operand = tokens[-1]
            if not any(isinstance(left_operand, t) for t in (Token, list, Literal)):
                raise ParseError("Invalid left operand", left_operand.start_position, token.start_position + len(token.input_text))
            tokens.pop()
            if not previous_tokens[i + 1:]:
                raise ParseError("Binary operator without right operand", token.start_position, token.start_position + len(token.input_text))
            right_operand = previous_tokens[i + 1]
            if not any(isinstance(right_operand, t) for t in (Token, list, Literal)):
                raise ParseError("Invalid right operand", token.start_position, right_operand.start_position + len(right_operand.input_text))
            skip_next_token = True
            expression = [left_operand, token, right_operand]
            tokens.append(expression)
    return tokens


def apply_unary_operator(tokens: typing.List[typing.Union[Token, Text, Operator, list]], operator: str) -> typing.List[typing.Union[Token, Text, Operator, list]]:
    previous_tokens = tokens
    tokens = []
    for i, token in reversed(list(enumerate(previous_tokens))):
        if isinstance(token, list):
            tokens.insert(0, apply_unary_operator(token, operator))
            continue
        if not isinstance(token, Operator):
            tokens.insert(0, token)
            continue
        if token.operator != operator:
            tokens.insert(0, token)
            continue
        if token.operator == operator:
            if not tokens:
                raise ParseError("Unary operator without operand", token.start_position, token.start_position + len(token.input_text))
            right_operand = tokens[0]
            if not any(isinstance(right_operand, t) for t in (Token, list, Literal)):
                raise ParseError("Invalid right operand", token.start_position, right_operand.start_position + len(right_operand.input_text))
            expression = [token, right_operand]
            tokens[0] = expression
    return tokens


def remove_redundant_lists(tokens: typing.List[typing.Union[Token, Text, Operator, list]]) -> typing.List[typing.Union[Token, Text, Operator, list]]:
    if len(tokens) == 1:
        if isinstance(tokens[0], list):
            return remove_redundant_lists(tokens[0])
        return tokens[0]
    previous_tokens = tokens
    tokens = []
    for token in previous_tokens:
        if isinstance(token, list):
            tokens.append(remove_redundant_lists(token))
            continue
        tokens.append(token)
    return tokens


def parse_date(text: str) -> typing.Optional[datatypes.DateTime]:
    for datetime_format in ['%Y-%m-%d', '%m/%d/%Y', '%d.%m.%Y']:
        try:
            return datatypes.DateTime(datetime.datetime.strptime(text.strip(), datetime_format))
        except ValueError:
            pass
    return None


def parse_quantity(text: str, start: int, end: int) -> typing.Optional[datatypes.Quantity]:
    had_decimal_point = False
    len_magnitude = 0
    start += text.find(text.lstrip())
    text = text.strip()
    end = start + len(text)
    is_negative = False

    for index, character in enumerate(text):
        if index == 0 and character == '-':
            is_negative = True
            continue
        if not character.isdigit():
            if not had_decimal_point and character == ".":
                had_decimal_point = True
            else:
                len_magnitude = index
                break
        else:
            len_magnitude = len(text)
    if is_negative:
        text = text[1:]
        len_magnitude -= 1

    if len_magnitude > 0:
        try:
            magnitude = int(text[:len_magnitude])
        except ValueError:
            magnitude = float(text[:len_magnitude])
        if is_negative:
            magnitude = -magnitude
        units = text[len_magnitude:]
        units = units.strip()
        if not units:
            units = None
        try:
            return datatypes.Quantity(magnitude, units)
        except ValueError:
            raise ParseError("Unable to parse units", start + len_magnitude, end)
    return None


def parse_bool(text: str) -> typing.Optional[datatypes.Boolean]:
    text = text.lower().strip()
    if text == 'true':
        return datatypes.Boolean(True)
    if text == 'false':
        return datatypes.Boolean(False)
    return None


def parse_tag(text: str) -> typing.Optional[str]:
    text = text.strip()
    if text.startswith('#'):
        return text[1:]
    return None


def parse_attribute(text: str, start: int, end: int) -> typing.Optional[typing.List[str]]:
    if text.strip()[:1] not in 'abcdefghijklmnopqrstuvwxyz':
        return None
    attributes = text.lower().strip().split('.')
    for attribute in attributes:
        if not all(character in 'abcdefghijklmnopqrstuvwxyz0123456789_?' for character in attribute):
            raise ParseError("Invalid attribute name", start, end)
        if '?' in attribute and attribute != '?':
            raise ParseError("Invalid array placeholder", start, end)
    if attributes.count('?') > 1:
        raise ParseError("Multiple array placeholders", start, end)
    return attributes


def convert_literals(tokens: typing.List[typing.Union[Token, Text, Operator, list]]) -> typing.List[typing.Union[Operator, Literal, list]]:
    previous_tokens = tokens
    tokens = []
    for token in previous_tokens:
        start = token.start_position
        end = token.start_position + len(token.input_text)
        if not isinstance(token, Token):
            tokens.append(token)
            continue
        tag = parse_tag(token.input_text)
        if tag is not None:
            accepted_characters = 'abcdefghijklmnopqrstuvwxyz0123456789_-äöüß'
            if not tag or not all(c in accepted_characters for c in tag):
                raise ParseError("Invalid tag", start + 1, end)
            tokens.append(Tag(token.input_text, token.start_position, tag))
            continue
        boolean = parse_bool(token.input_text)
        if boolean is not None:
            tokens.append(Boolean(token.input_text, token.start_position, boolean))
            continue
        date = parse_date(token.input_text)
        if date is not None:
            tokens.append(Date(token.input_text, token.start_position, date))
            continue
        quantity = parse_quantity(token.input_text, start, end)
        if quantity is not None:
            tokens.append(Quantity(token.input_text, token.start_position, quantity))
            continue
        attributes = parse_attribute(token.input_text, start, end)
        if attributes is not None:
            tokens.append(Attribute(token.input_text, token.start_position, attributes))
            continue
        raise ParseError("Unable to parse literal", start, end)
    return tokens


def parse_query_string(text: str) -> typing.List[typing.Union[Operator, Literal, list]]:
    tokens = [Token(text, 0)]
    tokens = split_by_texts(tokens)
    tokens = split_by_operators(tokens, ['(', ')', ' in ', '==', '!=', '<=', '>=', '=', '<', '>', ' after ', ' on ', ' before ', '!', ' not ', ' and ', ' or ', ' && ', ' || '])
    tokens = convert_literals(tokens)
    tokens = apply_parentheses(tokens)
    for operator in ['in', '==', '!=', '>=', '<=', '>', '<', 'after', 'on', 'before']:
        tokens = apply_binary_operator(tokens, operator)
    for operator in ['not', '!']:
        tokens = apply_unary_operator(tokens, operator)
    for operator in ['and', '&&', 'or', '||']:
        tokens = apply_binary_operator(tokens, operator)
    tokens = remove_redundant_lists(tokens)
    return tokens



# coding: utf-8
"""

"""

import datetime
import typing
from ..models import APILogEntry, HTTPMethod
from .. import db

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def get_api_log_entries(api_token_id: int) -> typing.List[APILogEntry]:
    """
    Return a list of API log entries for a given API token.

    :param api_token_id: the ID of an existing API token
    :return: a list of API log entries for this token
    """
    return APILogEntry.query.filter_by(api_token_id=api_token_id).order_by(db.desc(APILogEntry.utc_datetime)).all()


def create_log_entry(api_token_id: int, method: HTTPMethod, route: str):
    """
    Create a new API log entry.

    :param api_token_id: the ID of an existing API token
    :param method: the HTTP method
    :param route: the route
    """
    api_log_entry = APILogEntry(
        api_token_id=api_token_id,
        method=method,
        route=route,
        utc_datetime=datetime.datetime.utcnow()
    )
    db.session.add(api_log_entry)
    db.session.commit()



# coding: utf-8
"""

"""

import logging

from .utils import find_migrations, should_skip_by_index, update_migration_index


def run(db):
    logger = logging.getLogger('sampledb.migrations')
    for index, name, function in find_migrations():
        logger.info('Migration #{} "{}":'.format(index, name))

        # Skip migration by migration index
        if should_skip_by_index(db, index):
            logger.info("Skipped (index).")
            continue

        try:
            # Perform migration
            if function(db):
                logger.info("Done.")
            else:
                logger.info("Skipped (condition).")

            # Update migration index to skip this migration by index in the future
            update_migration_index(db, index)
            db.session.commit()
        except Exception:
            db.session.rollback()
            raise



# coding: utf-8
"""
Implementation of generate_placeholder(schema)
"""

import typing

from ..errors import UndefinedUnitError, SchemaError
from .utils import get_dimensionality_for_units


def generate_placeholder(schema: dict, path: typing.Optional[typing.List[str]] = None) -> typing.Union[dict, list, None]:
    """
    Generates a placeholder object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default
    :raise SchemaError: if the type is missing or invalid, or if the schema
        contains a quantity with invalid units
    """
    if path is None:
        path = []
    if 'type' not in schema:
        raise SchemaError('invalid schema', path)
    if schema['type'] == 'array':
        return _generate_array_placeholder(schema, path)
    elif schema['type'] == 'object':
        return _generate_object_placeholder(schema, path)
    elif schema['type'] == 'text':
        return _generate_text_placeholder(schema, path)
    elif schema['type'] == 'datetime':
        return _generate_datetime_placeholder(schema, path)
    elif schema['type'] == 'bool':
        return _generate_bool_placeholder(schema, path)
    elif schema['type'] == 'quantity':
        return _generate_quantity_placeholder(schema, path)
    elif schema['type'] == 'sample':
        return _generate_sample_placeholder(schema, path)
    elif schema['type'] == 'measurement':
        return _generate_measurement_placeholder(schema, path)
    elif schema['type'] == 'object_reference':
        return _generate_object_reference_placeholder(schema, path)
    elif schema['type'] == 'tags':
        return _generate_tags_placeholder(schema, path)
    elif schema['type'] == 'hazards':
        return _generate_hazards_placeholder(schema, path)
    elif schema['type'] == 'user':
        return _generate_user_placeholder(schema, path)
    else:
        raise SchemaError('invalid type', path)


def _generate_array_placeholder(schema: dict, path: typing.List[str]) -> list:
    """
    Generates a placeholder array object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    if 'default' in schema:
        return schema['default']
    if 'minItems' not in schema:
        return []
    min_items = schema['minItems']
    item_schema = schema['items']
    return [
        generate_placeholder(item_schema, path + ['[?]'])
        for _ in range(min_items)
    ]


def _generate_tags_placeholder(schema: dict, path: typing.List[str]) -> dict:
    """
    Generates a placeholder tags object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    if 'default' in schema:
        return {
            '_type': 'tags',
            'tags': schema['default']
        }
    return {
        '_type': 'tags',
        'tags': []
    }


def _generate_hazards_placeholder(schema: dict, path: typing.List[str]) -> dict:
    """
    Generate a placeholder GHS hazards object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    return {
        '_type': 'hazards'
    }


def _generate_object_placeholder(schema: dict, path: typing.List[str]) -> dict:
    """
    Generates a placeholder object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object
    """
    if 'default' in schema:
        return schema['default']
    properties = schema['properties']
    return {
        property_name: generate_placeholder(property_schema, path + [property_name])
        for property_name, property_schema in properties.items()
    }


def _generate_bool_placeholder(schema: dict, path: typing.List[str]) -> typing.Union[dict, None]:
    """
    Generates a placeholder boolean object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default value
    """
    if 'default' not in schema:
        return None
    value = schema['default']
    return {
        '_type': 'bool',
        'value': value
    }


def _generate_text_placeholder(schema: dict, path: typing.List[str]) -> typing.Union[dict, None]:
    """
    Generates a placeholder text object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default text
    """
    if 'default' not in schema:
        return None
    text = schema['default']
    return {
        '_type': 'text',
        'text': text
    }


def _generate_datetime_placeholder(schema: dict, path: typing.List[str]) -> typing.Union[dict, None]:
    """
    Generates a placeholder datetime object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default datetime
    """
    if 'default' not in schema:
        return None
    utc_datetime = schema['default']
    return {
        '_type': 'datetime',
        'utc_datetime': utc_datetime
    }


def _generate_quantity_placeholder(schema: dict, path: typing.List[str]) -> typing.Union[dict, None]:
    """
    Generates a placeholder quantity object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: the generated object or None, if there is no default magnitude (in base units)
    :raise SchemaError: if the units are invalid
    """
    if 'default' not in schema:
        return None
    magnitude_in_base_units = schema['default']
    units = schema['units']

    try:
        dimensionality = get_dimensionality_for_units(units)
    except UndefinedUnitError:
        raise SchemaError('invalid units', path)
    return {
        '_type': 'quantity',
        'dimensionality': dimensionality,
        'units': units,
        'magnitude_in_base_units': magnitude_in_base_units
    }


def _generate_sample_placeholder(schema: dict, path: typing.List[str]) -> None:
    """
    Generates a placeholder sample object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there can be no default sample
    """
    return _generate_object_reference_placeholder(schema, path)


def _generate_measurement_placeholder(schema: dict, path: typing.List[str]) -> None:
    """
    Generates a placeholder measurement object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there can be no default measurement
    """
    return _generate_object_reference_placeholder(schema, path)


def _generate_object_reference_placeholder(schema: dict, path: typing.List[str]) -> None:
    """
    Generates a placeholder object reference object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there can be no default object reference
    """
    return None


def _generate_user_placeholder(schema: dict, path: typing.List[str]) -> None:
    """
    Generates a placeholder user object based on an object schema.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :return: None, as there can be no default user
    """
    return None



# coding: utf-8
"""

"""

import smtplib
import flask
import flask_mail

from .. import mail, db
from .security_tokens import generate_token
from ..models import Authentication, AuthenticationType, User


def send_user_invitation_email(email, invitation_id):
    token_data = {
        'email': email,
        'invitation_id': invitation_id
    }
    token = generate_token(
        token_data,
        salt='invitation',
        secret_key=flask.current_app.config['SECRET_KEY']
    )
    confirm_url = flask.url_for('frontend.invitation_route', token=token, _external=True)

    service_name = flask.current_app.config['SERVICE_NAME']
    subject = service_name + " Invitation"
    html = flask.render_template('mails/user_invitation.html', confirm_url=confirm_url)
    text = flask.render_template('mails/user_invitation.txt', confirm_url=confirm_url)
    while '\n\n\n' in text:
        text = text.replace('\n\n\n', '\n\n')
    try:
        mail.send(flask_mail.Message(
            subject,
            sender=flask.current_app.config['MAIL_SENDER'],
            recipients=[email],
            body=text,
            html=html
        ))
    except smtplib.SMTPRecipientsRefused:
        pass


def send_email_confirmation_email(email, user_id, salt):
    token_data = {
        'email': email,
        'user_id': user_id
    }
    token = generate_token(
        token_data,
        salt=salt,
        secret_key=flask.current_app.config['SECRET_KEY']
    )
    confirm_url = flask.url_for("frontend.user_preferences", user_id=user_id, token=token, _external=True)

    service_name = flask.current_app.config['SERVICE_NAME']
    subject = service_name + " Email Confirmation"
    html = flask.render_template('mails/email_confirmation.html', confirm_url=confirm_url)
    text = flask.render_template('mails/email_confirmation.txt', confirm_url=confirm_url)
    while '\n\n\n' in text:
        text = text.replace('\n\n\n', '\n\n')
    try:
        mail.send(flask_mail.Message(
            subject,
            sender=flask.current_app.config['MAIL_SENDER'],
            recipients=[email],
            body=text,
            html=html
        ))
    except smtplib.SMTPRecipientsRefused:
        pass


def send_recovery_email(email):
    users = User.query.filter_by(email=email).all()
    email_authentication = Authentication.query.filter(db.and_(Authentication.login['login'].astext == email, Authentication.type == AuthenticationType.EMAIL)).first()
    if email_authentication is not None and email_authentication.user not in users:
        users.append(email_authentication.user)

    if not users:
        return

    password_reset_urls = {}
    for user in users:
        for authentication_method in user.authentication_methods:
            if authentication_method.type != AuthenticationType.LDAP:
                password_reset_urls[authentication_method] = build_confirm_url(authentication_method)

    service_name = flask.current_app.config['SERVICE_NAME']
    subject = service_name + " Account Recovery"
    html = flask.render_template('mails/account_recovery.html', email=email, users=users, password_reset_urls=password_reset_urls)
    text = flask.render_template('mails/account_recovery.txt', email=email, users=users, password_reset_urls=password_reset_urls)
    while '\n\n\n' in text:
        text = text.replace('\n\n\n', '\n\n')
    try:
        mail.send(flask_mail.Message(
            subject,
            sender=flask.current_app.config['MAIL_SENDER'],
            recipients=[email],
            body=text,
            html=html
        ))
    except smtplib.SMTPRecipientsRefused:
        pass


def build_confirm_url(authentication_method, salt='password'):
    assert authentication_method.type != AuthenticationType.LDAP

    user_id = authentication_method.user_id
    token = generate_token(authentication_method.id, salt=salt,
                           secret_key=flask.current_app.config['SECRET_KEY'])
    return flask.url_for("frontend.user_preferences", user_id=user_id, token=token, _external=True)



# coding: utf-8
"""
Logic module for files

Users with WRITE permissions can upload files for samples or measurements. These
files are read only, so to edit an already uploaded file the user will have to
re-upload it. Accordingly, this module only allows creating new files and reading
existing ones.

As multiple files with the same name may exist, a per-object file id is used as
prefix for the actual file names on disk. For each sample or measurement, at
most 10000 files may be uploaded. This limit is arbitrary, but it should be big
enough so that it will not be encountered in practice. As a result, the prefix
will always be 5 characters in length (0000_ to 9999_). The actual file name
may be up to 150 bytes in length (bytes, not characters, encoded as UTF-8).

The files are stored in a directory named after the object's ID, which in turn
will be inside folders named after the action's ID.
"""

import collections
import datetime
import os
import typing

from .. import db
from . import user_log, object_log, objects, users
from ..models import files
from ..models.file_log import FileLogEntry, FileLogEntryType
from .errors import FileNameTooLongError, TooManyFilesForObjectError, FileDoesNotExistError, InvalidFileStorageError

FILE_STORAGE_PATH: str = None
MAX_NUM_FILES: int = 10000


class File(collections.namedtuple('File', ['id', 'object_id', 'user_id', 'utc_datetime', 'data'])):
    """
    This class provides an immutable wrapper around models.files.File.
    """

    def __new__(cls, id: int, object_id: int, user_id: int, utc_datetime: typing.Optional[datetime.datetime] = None, data: typing.Optional[typing.Dict[str, typing.Any]] = None):
        self = super(File, cls).__new__(cls, id, object_id, user_id, utc_datetime, data)
        self._is_hidden = None
        self._hide_reason = None
        self._title = None
        self._description = None
        return self

    @classmethod
    def from_database(cls, file: files.File) -> 'File':
        if file.data is not None:
            data = file.data
        else:
            data = {
                'storage': 'local',
                'original_file_name': file.original_file_name
            }
        return File(
            id=file.id,
            object_id=file.object_id,
            user_id=file.user_id,
            utc_datetime=file.utc_datetime,
            data=data)

    @property
    def storage(self) -> str:
        return self.data.get('storage', 'local')

    @property
    def original_file_name(self):
        if self.storage == 'local':
            return self.data['original_file_name']
        else:
            raise InvalidFileStorageError()

    @property
    def real_file_name(self) -> str:
        if self.storage == 'local':
            # ensure that 4 digits are enough for every valid file ID
            assert MAX_NUM_FILES <= 10000
            # TODO: make the base path configurable
            object = objects.get_object(self.object_id)
            action_id = object.action_id
            prefixed_file_name = '{file_id:04d}_{file_name}'.format(file_id=self.id, file_name=self.original_file_name)
            return os.path.join(FILE_STORAGE_PATH, str(action_id), str(self.object_id), prefixed_file_name)
        else:
            raise InvalidFileStorageError()

    @property
    def url(self) -> str:
        if self.storage == 'url':
            return self.data['url']
        else:
            raise InvalidFileStorageError()

    @property
    def uploader(self) -> users.User:
        return users.get_user(self.user_id)

    @property
    def title(self) -> typing.Union[str, None]:
        if self._title is None:
            log_entry = FileLogEntry.query.filter_by(
                object_id=self.object_id,
                file_id=self.id,
                type=FileLogEntryType.EDIT_TITLE
            ).order_by(FileLogEntry.utc_datetime.desc()).first()
            if log_entry is None:
                if self.storage == 'local':
                    return self.original_file_name
                elif self.storage == 'url':
                    return self.data['url']
                else:
                    raise InvalidFileStorageError()
            self._title = log_entry.data['title']
        return self._title

    @property
    def description(self) -> typing.Union[str, None]:
        if self._description is None:
            log_entry = FileLogEntry.query.filter_by(
                object_id=self.object_id,
                file_id=self.id,
                type=FileLogEntryType.EDIT_DESCRIPTION
            ).order_by(FileLogEntry.utc_datetime.desc()).first()
            if log_entry is None:
                return None
            self._description = log_entry.data['description']
        return self._description

    @property
    def log_entries(self):
        return FileLogEntry.query.filter_by(
            object_id=self.object_id,
            file_id=self.id
        ).order_by(FileLogEntry.utc_datetime.asc()).all()

    @property
    def is_hidden(self):
        if self._is_hidden is None:
            log_entry = FileLogEntry.query.filter_by(
                object_id=self.object_id,
                file_id=self.id,
                type=FileLogEntryType.HIDE_FILE
            ).order_by(FileLogEntry.utc_datetime.desc()).first()
            if log_entry is None:
                self._is_hidden = False
                return False
            hide_time = log_entry.utc_datetime
            self._hide_reason = log_entry.data['reason']
            log_entry = FileLogEntry.query.filter_by(
                object_id=self.object_id,
                file_id=self.id,
                type=FileLogEntryType.UNHIDE_FILE
            ).order_by(FileLogEntry.utc_datetime.desc()).first()
            if log_entry is None:
                self._is_hidden = True
                return True
            unhide_time = log_entry.utc_datetime
            self._is_hidden = hide_time > unhide_time
        return self._is_hidden

    @property
    def hide_reason(self):
        if self.is_hidden:
            return self._hide_reason
        return None

    def open(self, read_only: bool = True) -> typing.BinaryIO:
        if self.storage == 'local':
            file_name = self.real_file_name
            if read_only:
                mode = 'rb'
            else:
                # before creating the file, the parent directories need exist
                os.makedirs(os.path.dirname(file_name), exist_ok=True)
                mode = 'xb'
            return open(file_name, mode)
        else:
            raise InvalidFileStorageError()


def create_local_file(object_id: int, user_id: int, file_name: str, save_content: typing.Callable[[typing.BinaryIO], None]) -> File:
    """
    Create a new local file and add it to the object and user logs.

    The function will call save_content with the opened file (in binary mode).

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param file_name: the original file name
    :param save_content: a function which will save the file's content to the
        given stream. The function will be called at most once.
    :return: the newly created file
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.FileNameTooLongError: when the file name is longer than 150
        bytes when encoded as UTF-8
    :raise errors.TooManyFilesForObjectError: when there are already 10000
        files for the object with the given id
    """
    # ensure that the file name is valid
    if len(file_name.encode('utf8')) > 150:
        raise FileNameTooLongError()

    db_file = _create_db_file(
        object_id=object_id,
        user_id=user_id,
        data={
            'storage': 'local',
            'original_file_name': file_name
        }
    )
    file = File.from_database(db_file)
    try:
        with file.open(read_only=False) as storage_file:
            save_content(storage_file)
    except Exception:
        db.session.delete(db_file)
        db.session.commit()
        raise
    _create_file_logs(file)
    return file


def create_url_file(object_id: int, user_id: int, url: str) -> File:
    """
    Create a file as a link to a URL and add it to the object and user logs.

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param url: the file URL
    :return: the newly created file
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.TooManyFilesForObjectError: when there are already 10000
        files for the object with the given id
    """
    db_file = _create_db_file(
        object_id=object_id,
        user_id=user_id,
        data={
            'storage': 'url',
            'url': url
        }
    )
    file = File.from_database(db_file)
    _create_file_logs(file)
    return file


def _create_file_logs(file: File) -> None:
    """
    Create object and user logs for a new file.

    :param file: the file to create log entries for
    """
    file_id = file.id
    object_id = file.object_id
    user_id = file.user_id
    object_log.upload_file(user_id=user_id, object_id=object_id, file_id=file_id)
    user_log.upload_file(user_id=user_id, object_id=object_id, file_id=file_id)


def _create_db_file(object_id: int, user_id: int, data: typing.Dict[str, typing.Any]) -> files.File:
    """
    Creates a new file and adds it to the object and user logs. The function
    will call save_content with the opened file (in binary mode).

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param data: the file storage data
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.TooManyFilesForObjectError: when there are already 10000
        files for the object with the given id
    """
    # ensure that the object exists
    object = objects.get_object(object_id)
    # ensure that the user exists
    users.get_user(user_id)
    # calculate the next file id
    previous_file_id = db.session.query(db.func.max(files.File.id)).filter(files.File.object_id == object.id).scalar()
    if previous_file_id is None:
        file_id = 0
    else:
        file_id = previous_file_id + 1
    if file_id >= MAX_NUM_FILES:
        raise TooManyFilesForObjectError()
    db_file = files.File(
        file_id=file_id,
        object_id=object_id,
        user_id=user_id,
        data=data
    )
    db.session.add(db_file)
    db.session.commit()
    return db_file


def update_file_information(object_id: int, file_id: int, user_id: int, title: str, description: str) -> None:
    """
    Creates new file log entries for updating a file's information.

    :param object_id: the ID of an existing object
    :param file_id: the ID of a file for the object
    :param user_id: the ID of an existing user
    :param title: the new title
    :param description: the new description
    :raise errors.FileDoesNotExistError: when no file with the given object ID
        and file ID exists
    """
    file = get_file_for_object(object_id=object_id, file_id=file_id)
    if file is None:
        raise FileDoesNotExistError()
    if not title:
        title = file.original_file_name
    if title != file.title:
        log_entry = FileLogEntry(type=FileLogEntryType.EDIT_TITLE, object_id=object_id, file_id=file_id, user_id=user_id, data={
            'title': title
        })
        db.session.add(log_entry)
        db.session.commit()
    if description != file.description and not (description == "" and file.description is None):
        log_entry = FileLogEntry(type=FileLogEntryType.EDIT_DESCRIPTION, object_id=object_id, file_id=file_id, user_id=user_id, data={
            'description': description
        })
        db.session.add(log_entry)
        db.session.commit()


def hide_file(object_id: int, file_id: int, user_id: int, reason: str) -> None:
    """
    Hides a file.

    :param object_id: the ID of an existing object
    :param file_id: the ID of a file for the object
    :param user_id: the ID of an existing user
    :param reason: the reason for hiding the file
    :raise errors.FileDoesNotExistError: when no file with the given object ID
        and file ID exists
    """
    file = get_file_for_object(object_id=object_id, file_id=file_id)
    if file is None:
        raise FileDoesNotExistError()
    if not reason:
        reason = ''
    log_entry = FileLogEntry(type=FileLogEntryType.HIDE_FILE, object_id=object_id, file_id=file_id, user_id=user_id, data={
        'reason': reason
    })
    db.session.add(log_entry)
    db.session.commit()


def get_file_for_object(object_id: int, file_id: int) -> typing.Optional[File]:
    """
    Returns the file with the given file ID for the object with the given object ID.

    :param object_id: the ID of an existing object
    :param file_id: the ID of an existing file for the object
    :return: the file or None
    """
    db_file = files.File.query.filter_by(object_id=object_id, id=file_id).first()
    if db_file is None:
        return None
    return File.from_database(db_file)


def get_files_for_object(object_id: int) -> typing.List[File]:
    """
    Returns a list of files for an object.

    :param object_id: the ID of an existing object
    :return: the list of files, sorted by upload time from first to last
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    db_files = files.File.query.filter_by(object_id=object_id).order_by(db.asc(files.File.utc_datetime)).all()
    if not db_files:
        # ensure that the object exists
        objects.get_object(object_id)
    return [File.from_database(db_file) for db_file in db_files]



# coding: utf-8
"""
RESTful API for SampleDB
"""

import base64
import json

import flask
from flask_restful import Resource

from .authentication import multi_auth
from ...logic.instruments import get_instrument
from ...logic.instrument_log_entries import get_instrument_log_entries, get_instrument_log_entry, get_instrument_log_file_attachment, get_instrument_log_object_attachment, get_instrument_log_categories, create_instrument_log_entry, create_instrument_log_file_attachment, create_instrument_log_object_attachment, InstrumentLogEntry, InstrumentLogEntryVersion, InstrumentLogFileAttachment, InstrumentLogObjectAttachment, InstrumentLogCategory
from ...logic import errors
from ...logic.object_permissions import get_user_object_permissions, Permissions

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

MAX_FILE_NAME_LENGTH = 256


def instrument_log_entry_to_json(log_entry: InstrumentLogEntry):
    return {
        'log_entry_id': log_entry.id,
        'author': log_entry.user_id,
        'versions': [
            instrument_log_entry_version_to_json(version)
            for version in log_entry.versions
        ]
    }


def instrument_log_entry_version_to_json(version: InstrumentLogEntryVersion):
    return {
        'log_entry_id': version.log_entry_id,
        'version_id': version.version_id,
        'utc_datetime': version.utc_datetime.isoformat(),
        'content': version.content,
        'categories': [
            category_to_json(category)
            for category in version.categories
        ],
    }


def file_attachment_to_json(file_attachment: InstrumentLogFileAttachment):
    return {
        'file_attachment_id': file_attachment.id,
        'file_name': file_attachment.file_name if not file_attachment.is_hidden else None,
        'base64_content': base64.b64encode(file_attachment.content).decode('ascii') if not file_attachment.is_hidden else None,
        'is_hidden': file_attachment.is_hidden
    }


def object_attachment_to_json(object_attachment: InstrumentLogObjectAttachment):
    return {
        'object_attachment_id': object_attachment.id,
        'object_id': object_attachment.object_id if not object_attachment.is_hidden else None,
        'is_hidden': object_attachment.is_hidden
    }


def category_to_json(category: InstrumentLogCategory):
    return {
        'category_id': category.id,
        'title': category.title
    }


class InstrumentLogEntries(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log entries for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        log_entries = get_instrument_log_entries(instrument_id)
        return [instrument_log_entry_to_json(log_entry) for log_entry in log_entries]

    @multi_auth.login_required
    def post(self, instrument_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_create_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log entries for instrument {instrument_id} can only be created by instrument scientists"
            }, 403
        try:
            data = json.loads(flask.request.data)
            unexpected_keys = set(data.keys()) - {'content', 'category_ids', 'file_attachments', 'object_attachments'}
            if unexpected_keys:
                raise ValueError()
            content = data['content']
            category_ids = data.get('category_ids', [])
            file_attachments = data.get('file_attachments', [])
            object_attachments = data.get('object_attachments', [])
        except Exception:
            return {
                "message": "expected json object containing content, category_ids, file_attachments and object attachments"
            }, 400
        if content is None:
            content = ''
        if not isinstance(content, str):
            return {
                "message": "expected string or null for content"
            }, 400
        try:
            if not isinstance(category_ids, list):
                raise TypeError()
            for category_id in category_ids:
                if not isinstance(category_id, int):
                    raise TypeError()
        except TypeError:
            return {
                "message": "expected list containing integer numbers for category_ids"
            }, 400

        existing_category_ids = {
            category.id
            for category in get_instrument_log_categories(instrument_id)
        }
        unknown_category_ids = set(category_ids) - existing_category_ids
        if unknown_category_ids:
            return {
                "message": f"unknown category_ids: {', '.join(map(str, unknown_category_ids))}"
            }, 400

        try:
            if not isinstance(file_attachments, list):
                raise TypeError()
            for file_attachment in file_attachments:
                if not isinstance(file_attachment, dict):
                    raise TypeError()
                if set(file_attachment.keys()) - {'file_name', 'base64_content'}:
                    raise ValueError()
                if not isinstance(file_attachment['file_name'], str):
                    raise TypeError()
                file_attachment['file_name'] = file_attachment['file_name'].strip()
                if not isinstance(file_attachment['base64_content'], str):
                    raise TypeError()
                file_attachment['content'] = base64.b64decode(file_attachment['base64_content'].encode('ascii'))
        except Exception:
            return {
                "message": "expected list containing dicts containing file_name and base64_content for file_attachments"
            }, 400

        for file_attachment in file_attachments:
            if not file_attachment['file_name'] or len(file_attachment['file_name']) > MAX_FILE_NAME_LENGTH:
                return {
                    "message": f"file attachment names must not be empty and must contain at most {MAX_FILE_NAME_LENGTH} characters"
                }, 400

        try:
            if not isinstance(object_attachments, list):
                raise TypeError()
            for object_attachment in object_attachments:
                if not isinstance(object_attachment, dict):
                    raise TypeError()
                if set(object_attachment.keys()) - {'object_id'}:
                    raise ValueError()
                if not isinstance(object_attachment['object_id'], int):
                    raise TypeError()
                if object_attachment['object_id'] < 0:
                    raise ValueError()
        except Exception:
            return {
                "message": "expected list containing dicts containing object_id for object_attachments"
            }, 400

        unknown_object_ids = set()
        attached_object_ids = []
        for object_attachment in object_attachments:
            object_id = object_attachment['object_id']
            try:
                permissions = get_user_object_permissions(object_id, flask.g.user.id)
                if Permissions.READ not in permissions:
                    raise errors.ObjectDoesNotExistError()
            except errors.ObjectDoesNotExistError:
                unknown_object_ids.add(object_id)
            else:
                if object_id in attached_object_ids:
                    return {
                        "message": f"duplicate object id in object_attachments: {object_id}"
                    }, 400
                attached_object_ids.append(object_id)
        if unknown_object_ids:
            return {
                "message": f"unknown object ids in object_attachments: {', '.join(map(str, unknown_object_ids))}"
            }, 400

        if not (content or file_attachments or object_attachments):
            return {
                "message": "log entry must contain content or an attachment"
            }, 400

        log_entry = create_instrument_log_entry(
            instrument_id=instrument.id,
            user_id=flask.g.user.id,
            content=content,
            category_ids=category_ids
        )

        for file_attachment in file_attachments:
            create_instrument_log_file_attachment(
                instrument_log_entry_id=log_entry.id,
                file_name=file_attachment['file_name'],
                content=file_attachment['content']
            )

        for object_id in attached_object_ids:
            create_instrument_log_object_attachment(
                instrument_log_entry_id=log_entry.id,
                object_id=object_id
            )

        return None, 201, {
            'Location': flask.url_for(
                'api.instrument_log_entry',
                instrument_id=instrument_id,
                log_entry_id=log_entry.id,
                _external=True
            )
        }


class InstrumentLogEntry(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int, log_entry_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log entries for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        try:
            log_entry = get_instrument_log_entry(log_entry_id)
            if log_entry.instrument_id != instrument_id:
                raise errors.InstrumentLogEntryDoesNotExistError()
        except errors.InstrumentLogEntryDoesNotExistError:
            return {
                "message": f"log entry {log_entry_id} for instrument {instrument_id} does not exist"
            }, 404
        return instrument_log_entry_to_json(log_entry)


class InstrumentLogEntryFileAttachments(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int, log_entry_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log entries for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        try:
            log_entry = get_instrument_log_entry(log_entry_id)
            if log_entry.instrument_id != instrument_id:
                raise errors.InstrumentLogEntryDoesNotExistError()
        except errors.InstrumentLogEntryDoesNotExistError:
            return {
                "message": f"log entry {log_entry_id} for instrument {instrument_id} does not exist"
            }, 404
        return [
            file_attachment_to_json(file_attachment)
            for file_attachment in log_entry.file_attachments
        ]


class InstrumentLogEntryFileAttachment(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int, log_entry_id: int, file_attachment_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log entries for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        try:
            log_entry = get_instrument_log_entry(log_entry_id)
            if log_entry.instrument_id != instrument_id:
                raise errors.InstrumentLogEntryDoesNotExistError()
        except errors.InstrumentLogEntryDoesNotExistError:
            return {
                "message": f"log entry {log_entry_id} for instrument {instrument_id} does not exist"
            }, 404
        try:
            file_attachment = get_instrument_log_file_attachment(file_attachment_id)
            if file_attachment.log_entry_id != log_entry_id:
                raise errors.InstrumentLogFileAttachmentDoesNotExistError()
        except errors.InstrumentLogFileAttachmentDoesNotExistError:
            return {
                "message": f"file attachment {file_attachment_id} for log entry {log_entry_id} for instrument {instrument_id} does not exist"
            }, 404
        return file_attachment_to_json(file_attachment)


class InstrumentLogEntryObjectAttachments(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int, log_entry_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log entries for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        try:
            log_entry = get_instrument_log_entry(log_entry_id)
            if log_entry.instrument_id != instrument_id:
                raise errors.InstrumentLogEntryDoesNotExistError()
        except errors.InstrumentLogEntryDoesNotExistError:
            return {
                "message": f"log entry {log_entry_id} for instrument {instrument_id} does not exist"
            }, 404
        return [
            object_attachment_to_json(object_attachment)
            for object_attachment in log_entry.object_attachments
        ]


class InstrumentLogEntryObjectAttachment(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int, log_entry_id: int, object_attachment_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log entries for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        try:
            log_entry = get_instrument_log_entry(log_entry_id)
            if log_entry.instrument_id != instrument_id:
                raise errors.InstrumentLogEntryDoesNotExistError()
        except errors.InstrumentLogEntryDoesNotExistError:
            return {
                "message": f"log entry {log_entry_id} for instrument {instrument_id} does not exist"
            }, 404
        try:
            object_attachment = get_instrument_log_object_attachment(object_attachment_id)
            if object_attachment.log_entry_id != log_entry_id:
                raise errors.InstrumentLogObjectAttachmentDoesNotExistError()
        except errors.InstrumentLogObjectAttachmentDoesNotExistError:
            return {
                "message": f"object attachment {object_attachment_id} for log entry {log_entry_id} for instrument {instrument_id} does not exist"
            }, 404
        return object_attachment_to_json(object_attachment)


class InstrumentLogCategories(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log categories for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        categories = get_instrument_log_categories(instrument_id)
        return [
            category_to_json(category)
            for category in categories
        ]


class InstrumentLogCategory(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int, category_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        if not instrument.users_can_view_log_entries and flask.g.user not in instrument.responsible_users:
            return {
                "message": f"log categories for instrument {instrument_id} can only be accessed by instrument scientists"
            }, 403
        categories = get_instrument_log_categories(instrument_id)
        for category in categories:
            if category.id == category_id:
                return category_to_json(category)
        return {
            "message": f"log category {category_id} for instrument {instrument_id} does not exist"
        }, 404



# coding: utf-8
"""

"""

import datetime
import typing

from .. import db
from .objects import Objects


class File(db.Model):
    __tablename__ = 'files'

    id = db.Column(db.Integer, primary_key=True)
    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)
    data = db.Column(db.JSON, nullable=True)
    uploader = db.relationship('User')

    def __init__(self, file_id: int, object_id: int, user_id: int, utc_datetime: typing.Optional[datetime.datetime] = None, data: typing.Optional[typing.Dict[str, typing.Any]] = None):
        self.id = file_id
        self.object_id = object_id
        self.user_id = user_id
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        self.utc_datetime = utc_datetime
        self.data = data

    def __repr__(self):
        return '<{0}(id={1.id}, object_id={1.object_id}, user_id={1.user_id}, utc_datetime={1.utc_datetime}, data="{1.data}")>'.format(type(self).__name__, self)
